import numpy as np

from rock_physics_open.equinor_utilities.various_utilities.types import (
    Array1D,
    Array2D,
    Array3D,
    Array4D,
)

from .array_functions import array_inverse, array_matrix_mult
from .calc_t import calc_t_vec
from .calc_theta import calc_theta_vec
from .calc_z import calc_z_vec


def calc_c_eff_vec(
    c0: Array3D[np.float64],
    c1: Array3D[np.float64],
    gd: Array3D[np.float64],
    t: Array4D[np.complex128],
    t_bar: Array4D[np.complex128],
    v: Array2D[np.float64],
    frac_aniso: float,
) -> Array3D[np.float64]:
    """
    Equation  4 (page 222) Jakobsen et al. 2003 (The acoustic signature of fluid flow in complex
    porous media).
    Returns the effective stiffness tensor C* (nx6x6 matrix) calculated from the t-matrices t(r) (Eq. 1).

    Parameters
    ----------
    c0 : np.ndarray
        Stiffness tensor of the host of the inclusion (nx6x6 matrix).
    c1 : np.ndarray
        Sum of the concentration and t-matrices (nx6x6 matrix).
    gd : np.ndarray
        Correlation function (nx6x6 matrix).
    t : np.ndarray
        t-matrices of the different inclusions (nx6x6xr matrix) (anisotropic).
    t_bar: np.ndarray
        t-matrices of the different inclusions (nx6x6xr matrix) (isotropic).
    v : np.ndarray
        Concentration of the inclusions (nxr vector).
    frac_aniso: np.ndarray
        Fraction of anisotropic.

    Returns
    -------
    tuple
        c_eff: np.ndarray.
        c_eff: effective stiffness tensor C.
    """
    if not (
        c0.ndim == 3
        and c1.ndim == 3
        and gd.ndim == 3
        and t.ndim == 4
        and t_bar.ndim == 4
        and v.ndim == 2
        and np.all(
            np.array(
                [
                    c0.shape[1:2],
                    c1.shape[1:2],
                    gd.shape[1:2],
                    t.shape[1:2],
                    t_bar.shape[1:2],
                ]
            )
            == c0.shape[1]
        )
        and np.all(
            np.array(
                [
                    c0.shape[0],
                    c1.shape[0],
                    gd.shape[0],
                    t.shape[0],
                    t_bar.shape[0],
                    v.shape[0],
                ]
            )
            == c0.shape[0]
        )
    ):
        raise ValueError("calc_c_eff_vec: inconsistencies in input shapes")

    log_length = c0.shape[0]
    alpha_len = v.shape[1]
    v_ = v.reshape((log_length, 1, 1, alpha_len))
    i4 = np.eye(6)

    c1 = c1 + np.sum(frac_aniso * v_ * t + (1.0 - frac_aniso) * v_ * t_bar, axis=3)

    return c0 + array_matrix_mult(c1, array_inverse(i4 + array_matrix_mult(gd, c1)))


def calc_c_eff_visco_vec(
    vs: Array1D[np.float64],
    k_r: Array1D[np.float64],
    eta_f: Array1D[np.float64],
    v: Array2D[np.float64],
    gamma: Array2D[np.float64],
    tau: Array1D[np.float64],
    kd_uuvv: Array2D[np.float64],
    kappa: Array1D[np.float64],
    kappa_f: Array1D[np.float64],
    c0: Array3D[np.float64],
    s0: Array3D[np.float64],
    c1: Array3D[np.float64],
    td: Array4D[np.float64],
    td_bar: Array4D[np.float64],
    x: Array4D[np.float64],
    x_bar: Array4D[np.float64],
    gd: Array3D[np.float64],
    frequency: float,
    frac_ani: float,
) -> Array3D[np.float64]:
    """
    Returns the effective stiffness tensor C* for a visco-elastic system (6x6xnumber of frequencies).

    Parameters
    ----------
    vs : np.ndarray
        The velocity used to calculate the wave number.
    k_r : np.ndarray
        Klinkenberg permeability.
    eta_f : np.ndarray
        Viscosity (P).
    v : np.ndarray
        Concentration of the inclusions which are connected with respect to fluid flow.
    gamma : np.ndarray
        Gamma factor for each inclusion (1x(number of connected inclusions) vector).
    tau : float or np.ndarray
        Relaxation time constant.
    kd_uuvv : np.ndarray
        Kd_uuvv for each connected inclusion (1x(number of connected inclusions) vector.
    kappa : np.ndarray
        Bulk modulus of host material.
    kappa_f : np.ndarray
        Bulk modulus of the fluid.
    c0 : np.ndarray
        The stiffness tensor of host material (6x6 matrix).
    s0 : np.ndarray
        Inverse of C0.
    c1 : np.ndarray
        First order correction matrix(6x6 matrix). If there are isolated inclusions, C1 is sum of concentration and
        t-matrices of the isolated part of the porosity.
    td : np.ndarray
        t-matrix tensors.
    td_bar : np.ndarray
        t-matrices of the connected inclusions(6x6x(numbers of inclusions) matrix).
    x : np.ndarray
        X-tensor.
    x_bar : np.ndarray
        X-tensor of the connected inclusions (6x6x(numbers of inclusions) matrix).
    gd : np.ndarray
        Correlation function (6x6 matrix).
    frequency : float
        Frequency under consideration.
    frac_ani : np.ndarray
        Fraction of anisotropic inclusions.

    Returns
    -------
    np.ndarray
        Effective stiffness tensor.
    """
    dr = k_r / eta_f

    omega = 2 * np.pi * frequency
    k = omega / vs
    theta = calc_theta_vec(
        v=v,
        omega=omega,
        gamma=gamma,
        tau=tau,
        kd_uuvv=kd_uuvv,
        dr=dr,
        k=k,
        kappa=kappa,
        kappa_f=kappa_f,
    )
    z, z_bar = calc_z_vec(
        s0=s0,
        td=td,
        td_bar=td_bar,
        omega=omega,
        gamma=gamma,
        v=v * frac_ani,
        tau=tau,
    )
    t = calc_t_vec(
        td=td,
        theta=theta,
        x=x,
        z=z,
        omega=omega,
        gamma=gamma,
        tau=tau,
        k_fluid=kappa_f,
    )
    t_bar = calc_t_vec(
        td=td_bar,
        theta=theta,
        x=x_bar,
        z=z_bar,
        omega=omega,
        gamma=gamma,
        tau=tau,
        k_fluid=kappa_f,
    )

    return calc_c_eff_vec(
        c0=c0,
        c1=c1,
        gd=gd,
        t=t,
        t_bar=t_bar,
        v=v,
        frac_aniso=frac_ani,
    )



import sys

import numpy as np
import numpy.typing as npt
from scipy.integrate import odeint

from .pq import p_q_fcn


def dem_model(
    k1: npt.NDArray[np.float64],
    mu1: npt.NDArray[np.float64],
    rho1: npt.NDArray[np.float64],
    k2: npt.NDArray[np.float64],
    mu2: npt.NDArray[np.float64],
    rho2: npt.NDArray[np.float64],
    frac2: npt.NDArray[np.float64],
    asp2: npt.NDArray[np.float64],
    tol: float,
) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.float64], npt.NDArray[np.float64]]:
    """
    DEM - Effective elastic moduli using Differential Effective Medium formulation.

    Parameters
    ----------
    k1 : np.ndarray
        Bulk modulus of background matrix [Pa].
    mu1 : np.ndarray
        Shear modulus of background matrix [Pa].
    rho1 : np.ndarray
        Bulk density of background matrix [kg/m^3].
    k2 : np.ndarray
        Bulk modulus of inclusions [Pa].
    mu2 : np.ndarray
        Shear modulus of inclusions [Pa].
    rho2 : np.ndarray
        Bulk density of inclusions [kg/m^3].
    frac2 : np.ndarray
        Fraction of inclusions [fraction].
    asp2 : np.ndarray
        Aspect ratio of inclusions [ratio].
    tol: float
        Desired accuracy in the ODE solver.

    Returns
    -------
    tuple
        k, mu, rho : np.ndarray
        k: effective medium bulk modulus [Pa], mu: effective medium shear modulus [Pa], rho: bulk density [kg/m^3].

    Comments
    --------
    Written by T. Mukerji, SRB, Stanford University.
    Ported to Python by Harald Flesche, Equinor 2015.
    """

    # Make sure all log inputs are vectors, expand scalars, throw error if uneven
    # length and not scalars
    k = np.ones(k1.shape) * np.nan
    mu = np.ones(k1.shape) * np.nan
    rhob = rho1 * (1 - frac2) + rho2 * frac2

    # Check for the trivial cases: all inclusions or all matrix
    idx = np.logical_or(frac2 == 1.0, frac2 == 0.0)
    if np.any(idx):
        idx1 = frac2 == 1.0
        idx2 = frac2 == 0.0
        if np.any(idx1):
            k[idx1] = k2[idx1]
            mu[idx1] = mu2[idx1]
        if np.any(idx2):
            k[idx2] = k1[idx2]
            mu[idx2] = mu1[idx2]
    if np.all(idx):
        return k, mu, rhob

    # Need to run the DEM model for cases that are mixed
    idx1 = ~idx
    frac2 = frac2[idx1]

    # Only need to run the ODE on different initial conditions
    uni, idx_restore = np.unique(
        np.stack((k1[idx1], mu1[idx1], k2[idx1], mu2[idx1]), axis=1),
        return_inverse=True,
        axis=0,
    )
    k1, mu1, k2, mu2 = np.split(uni, 4, axis=1)
    # Must flatten arrays to avoid problems with dimensions
    k1 = k1.flatten()
    k2 = k2.flatten()
    mu1 = mu1.flatten()
    mu2 = mu2.flatten()

    k_vec = np.zeros(len(idx_restore))
    mu_vec = np.zeros(len(idx_restore))

    tot = k1.shape[0]

    # Must run ODE solver with values in increasing order, and must start at zero value
    i_x = np.argsort(frac2)
    frac2 = frac2[i_x]
    frac2 = np.insert(frac2, 0, 0)
    # Find reverse order
    i_y = np.argsort(i_x)

    # Too many data points create memory problems for ODE solver, set limit at lim = 1000
    lim = 1000
    runs = 1 if tot <= lim else tot // lim + 1
    for i in range(runs):
        start_i = i * lim
        end_i = min(tot, (i + 1) * lim)

        # Select part to run, cast to array in case it defaults to single value
        k1_tmp = np.array(k1[start_i:end_i])
        mu1_tmp = np.array(mu1[start_i:end_i])
        k2_tmp = np.array(k2[start_i:end_i])
        mu2_tmp = np.array(mu2[start_i:end_i])
        asp2_tmp = np.array(asp2[start_i:end_i])

        y0 = np.concatenate((k1_tmp, mu1_tmp))

        try:
            y_out = odeint(
                _demy_prime, y0, frac2, args=(k2_tmp, mu2_tmp, asp2_tmp), rtol=tol
            )
        except ValueError:
            raise ValueError(sys.exc_info())

        # Remove inserted zero-value in row 0, split output in k and mu
        # Pycharm reports wrong object type here, checked and found to be OK
        k_out, mu_out = np.split(y_out[1:, :], 2, axis=1)

        # Reorder rows back to original sequence of fraction values
        k_out = k_out[i_y, :]
        mu_out = mu_out[i_y, :]

        # Select correct column for the row values
        idx_pres_range = np.logical_and(idx_restore >= start_i, idx_restore < end_i)
        for j in range(len(idx_restore)):
            if idx_pres_range[j]:
                k_vec[j] = k_out[j, idx_restore[j] - start_i]
                mu_vec[j] = mu_out[j, idx_restore[j] - start_i]

    k[idx1] = k_vec
    mu[idx1] = mu_vec

    return k, mu, rhob


def _demy_prime(
    y: npt.NDArray[np.float64],
    t: float,
    k2: npt.NDArray[np.float64],
    mu2: npt.NDArray[np.float64],
    asp2: npt.NDArray[np.float64],
) -> npt.NDArray[np.float64]:
    """
    Used by DEM model.

    Parameters
    ----------
    y : np.ndarray
        y array.
    t : float
        t float.
    k2 : np.ndarray
        Bulk modulus of inclusions [Pa].
    mu2 : np.ndarray
        Shear modulus of inclusions [Pa].
    asp2 : np.ndarray
        Aspect ratio of inclusions [ratio].

    Returns
    -------
    np.ndarray
        k, mu array.

    Comments
    --------
    Written by T. Mukerji, Stanford University.
    Rewritten in Python by Harald Flesche, Equinor 2015.
    """
    # Input value vector consists of both k and mu values
    k, mu = np.split(y, 2)

    p, q = p_q_fcn(k, mu, k2, mu2, asp2)

    k_r_hs = (k2 - k) * p / (1 - t)
    mu_r_hs = (mu2 - mu) * q / (1 - t)

    return np.concatenate((k_r_hs, mu_r_hs))



import os

import numpy as np

from rock_physics_open.equinor_utilities.snapshot_test_utilities import (
    INITIATE,
    compare_snapshots,
    get_snapshot_name,
    read_snapshot,
    store_snapshot,
)
from rock_physics_open.sandstone_models import (
    cemented_shaly_sand_sandy_shale_model,
    friable_shaly_sand_sandy_shale_model,
)
from rock_physics_open.sandstone_models.friable_models import CoordinateNumberFunction

k_sst = 36.8e9 * np.ones(20)
mu_sst = 44.0e9 * np.ones(20)
rho_sst = 2650 * np.ones(20)
k_mud = 15.0e9 * np.ones(20)
mu_mud = 7.5e9 * np.ones(20)
rho_mud = 2680 * np.ones(20)
k_cem = 36.8e9 * np.ones(20)
mu_cem = 44.0e9 * np.ones(20)
rho_cem = 2650 * np.ones(20)
k_fl_sst = 0.8e9 * np.ones(20)
rho_fl_sst = 850 * np.ones(20)
k_fl_mud = 2.7e9 * np.ones(20)
rho_fl_mud = 1005 * np.ones(20)
phi = np.linspace(0.40, 0.07, 20)
p_eff_mud = 20.0e6 * np.ones(20)
p_eff_sst = 20.0e6 * np.ones(20)
shale_frac = np.linspace(0.0, 1.0, 20)
frac_cem = 0.05 * np.ones(20)
phi_c_sst = 0.45
phi_c_mud = 0.7
n_sst = 8.0
n_mud = 15.0
coord_num_func_mud: CoordinateNumberFunction = "PorBased"
coord_num_func_sst: CoordinateNumberFunction = "PorBased"
shear_red_sst = 0.25
shear_red_mud = 1.0
phi_intr_mud = 0.05


def test_cemented_sandy_shale_model():
    args = cemented_shaly_sand_sandy_shale_model(
        k_sst=k_sst,
        mu_sst=mu_sst,
        rho_sst=rho_sst,
        k_cem=k_cem,
        mu_cem=mu_cem,
        rho_cem=rho_cem,
        k_mud=k_mud,
        mu_mud=mu_mud,
        rho_mud=rho_mud,
        k_fl_sst=k_fl_sst,
        rho_fl_sst=rho_fl_sst,
        k_fl_mud=k_fl_mud,
        rho_fl_mud=rho_fl_mud,
        phi=phi,
        p_eff_mud=p_eff_mud,
        shale_frac=shale_frac,
        frac_cem=frac_cem,
        phi_c_sst=phi_c_sst,
        n_sst=n_sst,
        shear_red_sst=shear_red_sst,
        phi_c_mud=phi_c_mud,
        phi_intr_mud=phi_intr_mud,
        coord_num_func_mud=coord_num_func_mud,
        n_mud=n_mud,
        shear_red_mud=shear_red_mud,
    )

    if not os.path.isfile(get_snapshot_name()) or INITIATE:
        _ = store_snapshot(get_snapshot_name(), *args)
    else:
        assert compare_snapshots(args, read_snapshot(get_snapshot_name()))


def test_friable_sandy_shale_model():
    args = friable_shaly_sand_sandy_shale_model(
        k_sst=k_sst,
        mu_sst=mu_sst,
        rho_sst=rho_sst,
        k_mud=k_mud,
        mu_mud=mu_mud,
        rho_mud=rho_mud,
        k_fl_sst=k_fl_sst,
        rho_fl_sst=rho_fl_sst,
        k_fl_mud=k_fl_mud,
        rho_fl_mud=rho_fl_mud,
        phi=phi,
        p_eff_sst=p_eff_sst,
        p_eff_mud=p_eff_mud,
        shale_frac=shale_frac,
        phi_c_sst=phi_c_sst,
        phi_c_mud=phi_c_mud,
        phi_intr_mud=phi_intr_mud,
        coord_num_func_sst=coord_num_func_sst,
        n_sst=n_sst,
        coord_num_func_mud=coord_num_func_mud,
        n_mud=n_mud,
        shear_red_sst=shear_red_sst,
        shear_red_mud=shear_red_mud,
    )

    if not os.path.isfile(get_snapshot_name()) or INITIATE:
        _ = store_snapshot(get_snapshot_name(), *args)
    else:
        assert compare_snapshots(args, read_snapshot(get_snapshot_name()))



from typing import Callable, cast

import numpy as np
import numpy.typing as npt

from rock_physics_open.equinor_utilities import gen_utilities
from rock_physics_open.equinor_utilities.optimisation_utilities import (
    gen_opt_routine,
    opt_param_to_ascii,
    save_opt_params,
)
from rock_physics_open.sandstone_models.patchy_cement_model import (
    patchy_cement_model_weight,
)

from .curvefit_sandstone_models import (
    curvefit_patchy_cement,
)


def patchy_cement_model_optimisation(
    k_min: npt.NDArray[np.float64],
    mu_min: npt.NDArray[np.float64],
    rho_min: npt.NDArray[np.float64],
    k_cem: npt.NDArray[np.float64],
    mu_cem: npt.NDArray[np.float64],
    rho_cem: npt.NDArray[np.float64],
    k_fl: npt.NDArray[np.float64],
    rho_fl: npt.NDArray[np.float64],
    por: npt.NDArray[np.float64],
    p_eff: npt.NDArray[np.float64],
    vp: npt.NDArray[np.float64],
    vs: npt.NDArray[np.float64],
    rhob: npt.NDArray[np.float64],
    phi_c: float,
    file_out_str: str = "patchy_cement_optimal_params.pkl",
    display_results: bool = False,
    well_name: str = "Unknown well",
    opt_params_only: bool = False,
) -> (
    tuple[
        npt.NDArray[np.float64],
        npt.NDArray[np.float64],
        npt.NDArray[np.float64],
        npt.NDArray[np.float64],
        npt.NDArray[np.float64],
        npt.NDArray[np.float64],
        npt.NDArray[np.float64],
        npt.NDArray[np.float64],
    ]
    | npt.NDArray[np.float64]
):
    """Patchy cement model with optimisation for a selection of parameters.

    Parameters
    ----------
    k_min : np.ndarray
        Cement bulk modulus [Pa].
    mu_min : np.ndarray
        Cement shear modulus [Pa].
    rho_min : np.ndarray
        Cement density [kg/m^3].
    k_cem : np.ndarray
        Cement bulk modulus [Pa].
    mu_cem : np.ndarray
        Cement shear modulus [Pa].
    rho_cem : np.ndarray
        Cement density [kg/m^3].
    k_fl : np.ndarray
        Fluid bulk modulus [Pa].
    rho_fl : np.ndarray
        Fluid density [kg/m^3].
    por : np.ndarray
        Inclusion porosity [ratio].
    p_eff : np.ndarray
        Effective pressure log [Pa].
    vp : np.ndarray
        Compressional velocity log [m/s].
    vs : np.ndarray
        Shear velocity log [m/s].
    rhob : np.ndarray
        Bulk density log [kg/m^3].
    phi_c : float
        Critical porosity [fraction]
    file_out_str : str
        Output file name (string) to store optimal parameters (pickle format).
    display_results : bool
        Display optimal parameters in a window after run.
    well_name : str
        Name of well to be displayed in info box title.
    opt_params_only : bool
        return parameters from optimisation only
    Returns
    -------
    tuple
        vp_mod, vs_mod, rho_mod, ai_mod, vpvs_mod - modelled logs,
        vp_res, vs_res, rho_res - residual logs.
    """

    # Skip hardcoded Vp/Vs ratio
    def_vpvs = np.mean(vp / vs)
    # Set weight to vs to give vp and vs similar influence on optimisation
    y_data = np.stack([vp, vs * def_vpvs], axis=1)
    # Optimisation function for selected parameters
    opt_fun: Callable[..., npt.NDArray[np.float64]] = curvefit_patchy_cement
    # expand single value parameters to match logs length
    por, phi_c_, def_vpvs = cast(
        list[npt.NDArray[np.float64]],
        gen_utilities.dim_check_vector((por, phi_c, def_vpvs)),
    )
    x_data = np.stack(
        (
            k_min,
            mu_min,
            rho_min,
            k_cem,
            mu_cem,
            rho_cem,
            k_fl,
            rho_fl,
            por,
            p_eff,
            def_vpvs,
            phi_c_,
        ),
        axis=1,
    )

    # Params: weight_k, weight_mu, shear_red, frac_cem
    lower_bound = np.array(
        [
            0.0,  # weight_k
            0.0,  # weight_mu
            0.0,  # shear_red
            0.01,  # frac_cem
        ],
        dtype=float,
    )
    upper_bound = np.array(
        [
            1.0,  # weight_k
            1.0,  # weight_mu
            1.0,  # shear_red
            0.1,  # frac_cem
        ],
        dtype=float,
    )
    x0 = (upper_bound + lower_bound) / 2.0
    # Optimisation step without fluid substitution
    vel_mod, vel_res, opt_params = gen_opt_routine(
        opt_function=opt_fun,
        x_data_orig=x_data,
        y_data=y_data,
        x_init=x0,
        low_bound=lower_bound,
        high_bound=upper_bound,
    )

    # Reshape outputs and remove weight from vs
    vp_mod, vs_mod = [arr.flatten() for arr in np.split(vel_mod, 2, axis=1)]
    vp_res, vs_res = [arr.flatten() for arr in np.split(vel_res, 2, axis=1)]
    vs_mod = vs_mod / def_vpvs
    vs_res = vs_res / def_vpvs
    vpvs_mod = vp_mod / vs_mod
    # Calculate the modelled density
    # rho_cem??
    rhob_mod = rho_min * (1.0 - por) + por * rho_fl
    ai_mod = vp_mod * rhob_mod
    rhob_res = rhob_mod - rhob
    # Save the optimal parameters
    save_opt_params(
        opt_type="pat_cem",
        opt_params=opt_params,
        file_name=file_out_str,
        well_name=well_name,
    )
    if display_results:
        opt_param_to_ascii(in_file=file_out_str, well_name=well_name)
    if opt_params_only:
        return opt_params
    return vp_mod, vs_mod, rhob_mod, ai_mod, vpvs_mod, vp_res, vs_res, rhob_res


def patchy_cement_model_optimisation_multiwell(
    k_min: npt.NDArray[np.float64],
    mu_min: npt.NDArray[np.float64],
    rho_min: npt.NDArray[np.float64],
    k_cem: npt.NDArray[np.float64],
    mu_cem: npt.NDArray[np.float64],
    rho_cem: npt.NDArray[np.float64],
    k_fl: npt.NDArray[np.float64],
    rho_fl: npt.NDArray[np.float64],
    por: npt.NDArray[np.float64],
    p_eff: npt.NDArray[np.float64],
    vp: npt.NDArray[np.float64],
    vs: npt.NDArray[np.float64],
    rhob: npt.NDArray[np.float64],
    phi_c: float,
    file_out_str: str = "pat_cem.pkl",
    display_results: bool = False,
    well_name: str = "Unknown well",
) -> tuple[
    list[npt.NDArray[np.float64]],
    list[npt.NDArray[np.float64]],
    list[npt.NDArray[np.float64]],
    list[npt.NDArray[np.float64]],
    list[npt.NDArray[np.float64]],
    list[npt.NDArray[np.float64]],
    list[npt.NDArray[np.float64]],
    list[npt.NDArray[np.float64]],
]:
    # First: perform a calibration with all data, return optimal parameters only. Concatenate all list of arrays to
    # make a single set of input arrays
    k_min_all = np.concatenate(k_min, axis=0)
    mu_min_all = np.concatenate(mu_min, axis=0)
    rho_min_all = np.concatenate(rho_min, axis=0)
    k_cem_all = np.concatenate(k_cem, axis=0)
    mu_cem_all = np.concatenate(mu_cem, axis=0)
    rho_cem_all = np.concatenate(rho_cem, axis=0)
    k_fl_all = np.concatenate(k_fl, axis=0)
    rho_fl_all = np.concatenate(rho_fl, axis=0)
    por_all = np.concatenate(por, axis=0)
    p_eff_all = np.concatenate(p_eff, axis=0)
    vp_all = np.concatenate(vp, axis=0)
    vs_all = np.concatenate(vs, axis=0)
    rhob_all = np.concatenate(rhob, axis=0)

    opt_param = patchy_cement_model_optimisation(
        k_min=k_min_all,
        mu_min=mu_min_all,
        rho_min=rho_min_all,
        k_cem=k_cem_all,
        mu_cem=mu_cem_all,
        rho_cem=rho_cem_all,
        k_fl=k_fl_all,
        rho_fl=rho_fl_all,
        por=por_all,
        p_eff=p_eff_all,
        vp=vp_all,
        vs=vs_all,
        rhob=rhob_all,
        phi_c=phi_c,
        file_out_str=file_out_str,
        display_results=display_results,
        well_name=well_name,
        opt_params_only=True,
    )

    # Next: run patchy cement model for all wells using the optimal parameters
    weight_k = opt_param[0]
    weight_mu = opt_param[1]
    shear_red = opt_param[2]
    frac_cem = opt_param[3]
    vp_mod: list[npt.NDArray[np.float64]] = []
    vs_mod: list[npt.NDArray[np.float64]] = []
    rhob_mod: list[npt.NDArray[np.float64]] = []
    ai_mod: list[npt.NDArray[np.float64]] = []
    vpvs_mod: list[npt.NDArray[np.float64]] = []
    vp_res: list[npt.NDArray[np.float64]] = []
    vs_res: list[npt.NDArray[np.float64]] = []
    rhob_res: list[npt.NDArray[np.float64]] = []
    for i in range(len(k_min)):
        vp_tmp, vs_tmp, rhob_tmp, _, _ = patchy_cement_model_weight(
            k_min=k_min[i],
            mu_min=mu_min[i],
            rho_min=rho_min[i],
            k_cem=k_cem[i],
            mu_cem=mu_cem[i],
            rho_cem=rho_cem[i],
            k_fl=k_fl[i],
            rho_fl=rho_fl[i],
            phi=por[i],
            p_eff=p_eff[i],
            frac_cem=frac_cem,
            phi_c=phi_c,
            coord_num_func="PorBased",
            n=9,
            shear_red=shear_red,
            weight_k=weight_k,
            weight_mu=weight_mu,
        )
        vp_mod.append(vp_tmp)
        vs_mod.append(vs_tmp)
        rhob_mod.append(rhob_tmp)
        ai_mod.append(vp_tmp * rhob_tmp)
        vpvs_mod.append(vp_tmp / vs_tmp)
        vp_res.append(vp_tmp - vp[i])
        vs_res.append(vs_tmp - vs[i])
        rhob_res.append(rhob_tmp - rhob[i])

    return vp_mod, vs_mod, rhob_mod, ai_mod, vpvs_mod, vp_res, vs_res, rhob_res



import numpy as np
import numpy.typing as npt

from .post_prob import posterior_probability

NULL_CLASS = 0


def mahal_class(
    obs: npt.NDArray[np.float64],
    class_mean: npt.NDArray[np.float64],
    class_cov: npt.NDArray[np.float64],
    class_id: npt.NDArray[np.int64],
    thresh: float = np.inf,
):
    """
    Mahalanobis classification routine. All data points are assigned a class, unless a threshold is set

    Parameters
    ----------
    obs : np.ndarray
        An nxm array, where n is the number of samples and m is the number of variables.
    class_mean : np.ndarray
        A pxm array, where p is the number of classes and m is the number of variables.
    class_cov : np.ndarray
        A pxm array, where p is the number of classes and m is the number of variables.
    class_id : np.ndarray
        A p length vector, where p is the number of classes, containing class_id (integer numbers).
    thresh : float
        Unclassified threshold.

    Returns
    -------
    tuple
        mahal_class_arr, mahal_dist, mahal_pp : (np.ndarray, np.ndarray, np.ndarray).
        mahal_class_arr: nx1 vector. The classes are numbered 1 to m, and unclassified samples (with distance
        greater than thresh) are set to 0,
        mahal_dist:	nx1 vector with mahalanobis distance from the closest class centre to sample,
        mahal_pp: nx1 vector with posterior probability based on the distance to each class
    """

    # Find dimensions
    n = obs.shape[0]
    p = class_mean.shape[0]

    # Assign matrices
    dist = np.zeros((n, p))

    # Calculate distance for each class
    for i in range(p):
        cov_inv = np.linalg.inv(class_cov[:, :, i])
        delta = obs - class_mean[i, :]
        dist[:, i] = np.sqrt(np.einsum("nj,jk,nk->n", delta, cov_inv, delta))

    # Find the shortest distance, assign class, calculate posterior probability and
    # filter out observations with distance greater than the threshold
    mahal_class_arr = np.choose(np.argmin(dist, axis=1), class_id)
    mahal_dist = np.amin(dist, axis=1)
    mahal_pp = posterior_probability(mahal_dist, dist)
    d_idx = mahal_dist > thresh
    mahal_class_arr[d_idx] = NULL_CLASS

    return mahal_class_arr, mahal_dist, mahal_pp



import numpy as np

from rock_physics_open.equinor_utilities.gen_utilities import filter_output


class TestFilterOutput:
    def test_filter_output(self):
        a_inp = np.linspace(0, 1, 11)
        b_inp = np.linspace(0, 1, 11) * 2
        idx = np.ones(15).astype(bool)
        idx[6:8] = False
        idx[12:14] = False
        a_ref = np.ones_like(idx) * np.nan
        a_ref[idx] = a_inp
        b_ref = np.ones_like(idx) * np.nan
        b_ref[idx] = b_inp
        a_out, b_out = filter_output(idx, (a_inp, b_inp))
        np.testing.assert_equal(a_out, a_ref)
        np.testing.assert_equal(b_out, b_ref)



from typing import Literal

import numpy as np
import numpy.typing as npt


def _refl_models(
    vp: npt.NDArray[np.float64],
    vs: npt.NDArray[np.float64],
    rho: npt.NDArray[np.float64],
    theta: npt.NDArray[np.float64],
    k: npt.NDArray[np.float64] | float = 2.0,
    mod: Literal["aki_richards", "smith_gidlow"] = "aki_richards",
) -> npt.NDArray[np.float64]:
    """Calculate reflect coeff.

    Parameters
    ----------
    vp : np.ndarray
        vp array.
    vs : np.ndarray
        vs array.
    rho : np.ndarray
        rho array.
    theta : float, np.ndarray
        Theta value.
    k : float, np.ndarray, optional
        By default 2.0.
    mod : str, optional
        By default 'aki_richards'.

    Returns
    -------
    np.ndarray
        Reflect coeff.
    """
    theta = theta / 180 * np.pi
    r_vp = (vp[1:] - vp[0:-1]) / (vp[0:-1] + vp[1:])
    r_vs = (vs[1:] - vs[0:-1]) / (vs[0:-1] + vs[1:])
    r_rho = (rho[1:] - rho[0:-1]) / (rho[0:-1] + rho[1:])
    # Need to append one sample to make returned array of the same size - repeat the last value
    r_vp = np.append(r_vp, r_vp[-1])
    r_vs = np.append(r_vs, r_vs[-1])
    r_rho = np.append(r_rho, r_rho[-1])

    if mod == "aki_richards":
        reflect_coeff = (
            0.5 * (r_vp + r_rho)
            + (0.5 * r_vp - 2 * k**-2 * (2 * r_vs + r_rho)) * np.sin(theta) ** 2
            + 0.5 * r_vp * (np.tan(theta) ** 2 - np.sin(theta) ** 2)
        )
    else:
        reflect_coeff = (
            5 / 8 - 0.5 * k**-2 * np.sin(theta) ** 2 + 0.5 * np.tan(theta) ** 2
        ) * r_vp - 4 * k**-2 * np.sin(theta) ** 2 * r_vs

    return reflect_coeff


def aki_richards(
    vp: npt.NDArray[np.float64],
    vs: npt.NDArray[np.float64],
    rho: npt.NDArray[np.float64],
    theta: npt.NDArray[np.float64],
    k: npt.NDArray[np.float64] | float = 2.0,
) -> npt.NDArray[np.float64]:
    """
    Linearised Zoeppritz equation according to Aki and Richards.

    Parameters
    ----------
    vp : np.ndarray
        Pressure wave velocity [m/s].
    vs : np.ndarray
        Shear wave velocity [m/s].
    rho : np.ndarray
        Density [kg/m^3].
    theta : np.ndarray
        Angle of incident ray [radians].
    k : float, np.ndarray
        Background vp/vs [unitless].

    Returns
    -------
    refl_coeff : np.ndarray
        Reflection coefficient [unitless].
    """
    return _refl_models(vp, vs, rho, theta, k, mod="aki_richards")


def smith_gidlow(
    vp: npt.NDArray[np.float64],
    vs: npt.NDArray[np.float64],
    rho: npt.NDArray[np.float64],
    theta: npt.NDArray[np.float64],
    k: npt.NDArray[np.float64] | float = 2.0,
) -> npt.NDArray[np.float64]:
    """
    Linearised Zoeppritz equation according to Smith and Gidlow.

    Parameters
    ----------
    vp : np.ndarray
        Pressure wave velocity [m/s].
    vs : np.ndarray
        Shear wave velocity [m/s].
    rho : np.ndarray
        Density [kg/m^3].
    theta : np.ndarray
        Angle of incident ray [radians].
    k : float, np.ndarray
        Background vp/vs [unitless].

    Returns
    -------
    refl_coeff : np.ndarray
        Reflection coefficient [unitless].
    """

    return _refl_models(vp, vs, rho, theta, k, mod="smith_gidlow")



from .dummy_vars import generate_dummy_vars
from .exponential_model import ExponentialPressureModel
from .import_ml_models import import_model
from .polynomial_model import PolynomialPressureModel
from .run_regression import run_regression
from .sigmoidal_model import SigmoidalPressureModel

__all__ = [
    "generate_dummy_vars",
    "import_model",
    "run_regression",
    "ExponentialPressureModel",
    "PolynomialPressureModel",
    "SigmoidalPressureModel",
]



import numpy as np
import numpy.typing as npt


def live_oil_density(
    temperature: npt.NDArray[np.float64],
    pressure: npt.NDArray[np.float64] | None,  # pyright: ignore[reportUnusedParameter]
    reference_density: npt.NDArray[np.float64],
    gas_oil_ratio: npt.NDArray[np.float64],
    gas_gravity: npt.NDArray[np.float64],
) -> npt.NDArray[np.float64]:
    """
    Density of live oil at saturation.

    Equation 24 in Batzle & Wang [1].

    :param reference_density: Density of the oil without dissolved gas
        at 15.6 degrees Celsius and atmospheric pressure. [kg/m^3]
    :param pressure: Formation pressure [Pa] of oil (for future implementation only)
    :param gas_oil_ratio: The volume ratio of gas to oil [l/l]
    :param temperature: Temperature [°C] of oil.
    :param gas_gravity: molar mass of gas relative to air molar mas.
    :return: Density of live oil [kg/m^3].
    """
    density_gcc = reference_density / 1000.0
    b0 = live_oil_volume_factor(
        temperature=temperature,
        reference_density=reference_density,
        gas_oil_ratio=gas_oil_ratio,
        gas_gravity=gas_gravity,
    )
    return 1000.0 * (density_gcc + 0.0012 * gas_gravity * gas_oil_ratio) / b0


def live_oil_pseudo_density(
    temperature: npt.NDArray[np.float64],
    reference_density: npt.NDArray[np.float64],
    gas_oil_ratio: npt.NDArray[np.float64],
    gas_gravity: npt.NDArray[np.float64],
) -> npt.NDArray[np.float64]:
    """
    Pseudo density used to substitute reference density in dead_oil_wave_velocity
    for live oils.

    Equation 22 in Batzle & Wang [1].

    :param reference_density: Density of the oil without dissolved gas
        at 15.6 degrees Celsius and atmospheric pressure. [kg/m^3]
    :param gas_oil_ratio: The volume ratio of gas to oil [l/l]
    :param temperature: Temperature [°C] of oil.
    :param gas_gravity: molar mass of gas relative to air molar mas.
    :return: Pseudo-density of live oil [kg/m^3].
    """
    density_gcc = reference_density / 1000.0
    b0 = live_oil_volume_factor(
        temperature=temperature,
        reference_density=reference_density,
        gas_oil_ratio=gas_oil_ratio,
        gas_gravity=gas_gravity,
    )
    return 1000.0 * (density_gcc / b0) / (1 + 0.001 * gas_oil_ratio)


def live_oil_volume_factor(
    temperature: npt.NDArray[np.float64],
    reference_density: npt.NDArray[np.float64],
    gas_oil_ratio: npt.NDArray[np.float64],
    gas_gravity: npt.NDArray[np.float64],
) -> npt.NDArray[np.float64]:
    """
    Volume factor derived by Standing (1962), equation 23 in Batzle & Wang [1].
    :param reference_density: Density of the oil without dissolved gas
        at 15.6 degrees Celsius and atmospheric pressure. [kg/m^3]
    :param gas_oil_ratio: The volume ratio of gas to oil [l/l]
    :param temperature: Temperature [°C] of oil.
    :param gas_gravity: molar mass of gas relative to air molar mas.
    :return: A volume factor in calculating pseudo-density of live oil [unitless].
    """
    density_gcc = reference_density / 1000.0
    return (
        0.972
        + 0.00038
        * (
            2.4 * gas_oil_ratio * np.sqrt(gas_gravity / density_gcc)
            + temperature
            + 17.8
        )
        ** 1.175
    )



import os

import numpy as np

from rock_physics_open.equinor_utilities.snapshot_test_utilities import (
    INITIATE,
    compare_snapshots,
    get_snapshot_name,
    read_snapshot,
    store_snapshot,
)
from rock_physics_open.span_wagner import co2_properties

temp = 100.0 * np.linspace(0.8, 1.2, 101)
pres = 23.0e6 * np.linspace(0.8, 1.2, 101)
c1 = 0.1 * np.ones(101)
c2 = 0.1 * np.ones(101)
c3 = 0.1 * np.ones(101)
c4 = 0.1 * np.ones(101)
c5 = 0.1 * np.ones(101)
c6 = 0.1 * np.ones(101)
c7 = 0.1 * np.ones(101)
n2 = 0.1 * np.ones(101)
co2 = 0.05 * np.ones(101)
h2s = 0.05 * np.ones(101)
sgc7 = 0.81 * np.ones(101)
mwc7 = 161 * np.ones(101)
gr = 1.0 * np.linspace(0.7, 1.05, 101)


def test_co2_properties():
    temp_sw = np.linspace(-56, 28, 101)
    args = co2_properties(temp_sw, pres)

    if not os.path.isfile(get_snapshot_name()) or INITIATE:
        _ = store_snapshot(get_snapshot_name(), *args)
    else:
        assert compare_snapshots(args, read_snapshot(get_snapshot_name()))



from typing import Literal

import numpy as np

from rock_physics_open.equinor_utilities.various_utilities.types import (
    Array1D,
    Array2D,
    Array3D,
    Array4D,
)

from .calc_kd_eff import calc_kd_eff_vec


def calc_pressure_vec(
    alpha_con: Array2D[np.float64],
    alpha_iso: Array2D[np.float64],
    v_con: Array2D[np.float64],
    v_iso: Array2D[np.float64],
    c0: Array3D[np.float64],
    s_0: Array3D[np.float64],
    gd: Array3D[np.float64],
    d_p: float,
    tau: Array1D[np.float64],
    gamma: Array2D[np.float64],
    k_fl: Array1D[np.float64],
    ctrl: Literal[0, 1, 2],
    frac_ani: float,
) -> tuple[
    Array2D[np.float64],
    Array2D[np.float64],
    Array2D[np.float64],
    Array2D[np.float64],
    Array1D[np.float64],
    Array2D[np.float64],
]:
    """Calculate the effect of depletion on aspect ratios.

    Parameters
    ----------
    alpha_con : np.ndarray
        Aspect ratio for connected inclusions (r length vector).
    alpha_iso : np.ndarray
        Aspect ratio for connected inclusions (s length vector).
    v_con : np.ndarray
        Volume of connected inclusions (r length vector).
    v_iso : np.ndarray
        Volume of connected inclusions (s length vector).
    c0 : np.ndarray
        Stiffness tensor of host material (nx6x6 array).
    s_0 : np.ndarray
        Inverse of stiffness tensor (nx6x6 array).
    gd : np.ndarray
        The correlation function (green's tensor nx6x6 matrix).
    d_p : float
        Change in effective pressure.
    tau : np.ndarray
        Relaxation time constant ((numbers of connected pores) vector).
    gamma : np.ndarray
        Gamma factor ((numbers of connected pores) vector).
    k_fl : np.ndarray
        Fluid bulk modulus (n length vector).
    ctrl : int
        Control parameter.
    frac_ani : float
        Fraction of anisotropic inclusions.

    Returns
    -------
    tuple
        alpha_n_connected, v_n_connected, alpha_n_isolated, v_n_isolated, tau_n_out, gamma_n_out -
        modified alphas, volumes, relaxation times and gamma factors.

    Notes
    -----
    Equations used can be found in:
    Agersborg (2007), phd thesis:
    https://bora.uib.no/handle/1956/2422

    09.03.2012
    Remy Agersborg
    email: remy@agersborg.com

    Translated to Python and vectorised by Harald Flesche, hfle@equinor.com 2020
    """

    def _new_values(
        k: Array4D[np.float64],
        sum_k: Array2D[np.float64],
        a: Array2D[np.float64],
        v: Array2D[np.float64],
        d_p: float,
        t: Array1D[np.float64],
        g: Array2D[np.float64],
    ) -> tuple[
        Array2D[np.float64],
        Array2D[np.float64],
        Array1D[np.float64],
        Array2D[np.float64],
    ]:
        # Local helper function to avoid code duplication
        len_alpha = a.shape[1]
        len_log = k.shape[0]

        v_new = v * (
            1
            - (np.sum(k[:, 0:3, 0:3, :], axis=(1, 2)) - sum_k.reshape(len_log, 1)) * d_p
        )
        alpha_new = a * (
            1
            - (np.sum(k[:, 2, 0:3, :], axis=1) - np.sum(k[:, 0, 0:3, :], axis=1)) * d_p
        )

        idx_neg = (alpha_new < 0.0) | (v_new < 0.0)
        idx_high = (alpha_new > 1.0) | (v_new > 1.0)
        idx_inval = np.logical_or(idx_neg, idx_high)

        for i in range(len_alpha):
            # Maybe best to set undefined, but for now keep old values
            # alpha_new[idx_inval] = np.nan
            # v_new[idx_inval] = np.nan
            alpha_new[idx_inval[:, i], i] = a[idx_inval[:, i], i]
            v_new[idx_inval[:, i], i] = v[idx_inval[:, i], i]
        tau_n = np.array(t)
        gamma_n = np.array(g)

        return v_new, alpha_new, tau_n, gamma_n

    kd_eff_isolated, kd_eff_connected = calc_kd_eff_vec(
        c0=c0,
        s_0=s_0,
        k_fl=k_fl,
        alpha_con=alpha_con,
        alpha_iso=alpha_iso,
        v_con=v_con,
        v_iso=v_iso,
        gd=gd,
        ctrl=ctrl,
        frac_ani=frac_ani,
    )
    # Find the sum in the eq. 21 Jakobsen and Johansen 2005
    sum_kd = 0.0
    if ctrl != 2 and kd_eff_isolated is not None:
        count_isolated = kd_eff_isolated.shape[3]
        for j in range(count_isolated):
            sum_kd = sum_kd + v_iso[:, j] * np.sum(
                kd_eff_isolated[:, 0:3, 0:3, j], axis=(1, 2)
            )
    if ctrl != 0 and kd_eff_connected is not None:
        count_connected = kd_eff_connected.shape[3]
        for j in range(count_connected):
            sum_kd = sum_kd + v_con[:, j] * np.sum(
                kd_eff_connected[:, 0:3, 0:3, j], axis=(1, 2)
            )
    # Find the new concentration of inclusion
    alpha_n_isolated = None
    alpha_n_connected = None
    v_n_isolated = None
    v_n_connected = None
    gamma_n_out = None
    tau_n_out = None
    if ctrl != 2 and kd_eff_isolated is not None:
        v_n_isolated, alpha_n_isolated, _, _ = _new_values(
            k=kd_eff_isolated,
            sum_k=sum_kd,  # pyright: ignore[reportArgumentType] | sum_kd is Array2D after the loop
            a=alpha_iso,
            v=v_iso,
            d_p=d_p,
            t=tau,
            g=gamma,
        )

    if ctrl != 0 and kd_eff_connected is not None:
        v_n_connected, alpha_n_connected, tau_n_out, gamma_n_out = _new_values(
            k=kd_eff_connected,
            sum_k=sum_kd,  # pyright: ignore[reportArgumentType] | sum_kd is Array2D after the loop
            a=alpha_con,
            v=v_con,
            d_p=d_p,
            t=tau,
            g=gamma,
        )

    return (  # pyright: ignore[reportReturnType] | #TODO: This should be fixed properly
        alpha_n_connected,
        v_n_connected,
        alpha_n_isolated,
        v_n_isolated,
        tau_n_out,
        gamma_n_out,
    )



import warnings
from pickle import load

from sklearn.preprocessing import OneHotEncoder, RobustScaler

from .exponential_model import ExponentialPressureModel
from .polynomial_model import PolynomialPressureModel
from .sigmoidal_model import SigmoidalPressureModel

AnyModel = ExponentialPressureModel | PolynomialPressureModel | SigmoidalPressureModel


def import_model(
    model_file_name: str,
) -> tuple[
    AnyModel,
    RobustScaler,
    OneHotEncoder | None,
    str,
    str,
    list[str],
    str | list[str],
]:
    """
    Utility to import a pickled dict containing information needed to run a classification or regression based on
    a calibrated model.

    Parameters
    ----------
    model_file_name : str
        Full name including path for model file.

    Returns
    -------
    models, scaler, ohe, label_var, label_units, feature_var, cat_var : Any
        models: various regression or classification models from e.g. sklearn or tensorflow keras, scaler: preprocessing
        Robust Scaler, label_var: name(s) of label variable(s), label_unit: unit(s) of label variable(s), cat_var:
        categorical variables that should be encoded with one-hot-encoder.
    """

    with open(model_file_name, "rb") as fin, warnings.catch_warnings():
        # 11.04.2021 HFLE: There is an issue that is not connected to the local function, in that a warning is issued
        # when the model is loaded, claiming that it is of an older version. This is debugged in detail, and the model
        # IS of the correct version, so the error arise elsewhere. To avoid confusion, the warning is suppressed here

        warnings.simplefilter("ignore", category=UserWarning)
        mod_dict = load(fin)

    if mod_dict["model_type"] == "Sigmoid":
        models = SigmoidalPressureModel.load(mod_dict["nn_mod"])
    elif mod_dict["model_type"] == "Exponential":
        models = ExponentialPressureModel.load(mod_dict["nn_mod"])
    elif mod_dict["model_type"] == "Polynomial":
        models = PolynomialPressureModel.load(mod_dict["nn_mod"])
    else:
        raise ValueError("unknown model type {}".format(mod_dict["model_type"]))

    ohe: OneHotEncoder | None = None
    cat_var: str | list[str] = []
    try:
        if mod_dict["ohe"]:
            with open(mod_dict["ohe"], "rb") as f:
                ohe_dict = load(f)
                ohe = ohe_dict["ohe"]
                cat_var = ohe_dict["cat_var"]
    except (FileExistsError, FileNotFoundError):
        pass

    return (
        models,
        mod_dict["scaler"],
        ohe,
        mod_dict["label_var"],
        mod_dict["label_units"],
        mod_dict["feature_var"],
        cat_var,
    )



from warnings import warn

import numpy as np
import numpy.typing as npt


def rho_b(
    phi: npt.NDArray[np.float64],
    rho_f: npt.NDArray[np.float64],
    rho_mat: npt.NDArray[np.float64],
) -> npt.NDArray[np.float64]:
    """
    Calculate bulk density from porosity, fluid density and matrix density.

    Parameters
    ----------
    phi: np.ndarray
        Porosity [fraction].
    rho_f: np.ndarray
        Fluid bulk density [kg/m^3].
    rho_mat: np.ndarray
        Matrix bulk density [kg/m^3].

    Returns
    -------
    np.ndarray
        rhob: bulk density [kg/m^3].
    """
    return phi * rho_f + (1 - phi) * rho_mat


def rho_m(
    frac_cem: npt.NDArray[np.float64],
    phi: npt.NDArray[np.float64],
    rho_cem: npt.NDArray[np.float64],
    rho_min: npt.NDArray[np.float64],
) -> npt.NDArray[np.float64]:
    """
    Calculates matrix density as a combination of cement fraction and minerals
    fracCem is defined as cement fraction relative to total volume.

    Parameters
    ----------
    frac_cem : np.ndarray
        Cement fraction [fraction].
    phi : np.ndarray
        Porosity [fraction].
    rho_cem : np.ndarray
         Cement density [kg/m^3].
    rho_min : np.ndarray
         Mineral density [kg/m^3].

    Returns
    -------
    np.ndarray
        rho_mat: matrix density [kg/m^3]
    """
    idx = np.logical_and(phi >= 0.0, phi < 1.0)
    if np.sum(idx) != len(phi):
        warn(
            f"{__file__}: phi out of range in {len(phi) - np.sum(idx)} sample",
        )

    rho_mat = np.ones_like(phi) * np.nan
    f_cem = frac_cem[idx] / (1 - phi[idx])

    rho_mat[idx] = f_cem * rho_cem[idx] + (1 - f_cem) * rho_min[idx]

    return rho_mat



import numpy as np
import numpy.typing as npt


def p_q_fcn(
    k: npt.NDArray[np.float64],
    mu: npt.NDArray[np.float64],
    k2: npt.NDArray[np.float64],
    mu2: npt.NDArray[np.float64],
    asp: npt.NDArray[np.float64],
) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:
    """
    Geometric factors used in inclusion models.

    References
    ----------
    The rock physics handbook, Gary Mavko et al.

    Parameters
    ----------
    k : np.ndarray
        Bulk modulus of phase 1 [Pa].
    mu : np.ndarray
        Shear modulus of phase 1 [Pa].
    k2 : np.ndarray
        Bulk modulus of phase 2 [Pa].
    mu2 : np.ndarray
        Shear modulus of phase 2 [Pa].
    asp : np.ndarray
        Aspect ratio [ratio].

    Returns
    -------
    tuple
        p, q : np.ndarray
        geometric factors p and q.
    """

    # Functions theta and fn defaults to 2/3 and -2/5 for asp == 1.0
    idx_oblate = np.less(asp, 1.0)
    idx_prolate = np.greater(asp, 1.0)
    theta = 2.0 / 3.0 * np.ones(asp.shape)
    fn = -2.0 / 5.0 * np.ones(asp.shape)

    if np.any(idx_oblate):
        theta[idx_oblate] = (
            asp[idx_oblate] / ((1 - asp[idx_oblate] ** 2) ** (3 / 2))
        ) * (
            np.arccos(asp[idx_oblate])
            - asp[idx_oblate] * np.sqrt(1 - asp[idx_oblate] ** 2)
        )
        fn[idx_oblate] = (asp[idx_oblate] ** 2 / (1 - asp[idx_oblate] ** 2)) * (
            3 * theta[idx_oblate] - 2
        )

    if np.any(idx_prolate):
        theta[idx_prolate] = (
            asp[idx_prolate] / ((asp[idx_prolate] ** 2 - 1) ** (3 / 2))
        ) * (
            asp[idx_prolate] * np.sqrt(asp[idx_prolate] ** 2 - 1)
            - np.arccosh(asp[idx_prolate])
        )
        fn[idx_prolate] = (asp[idx_prolate] ** 2 / (asp[idx_prolate] ** 2 - 1)) * (
            2 - 3 * theta[idx_prolate]
        )

    nu = (3 * k - 2 * mu) / (2 * (3 * k + mu))
    r = (1 - 2 * nu) / (2 * (1 - nu))
    a = mu2 / mu - 1
    b = (1 / 3) * (k2 / k - mu2 / mu)

    f1 = 1 + a * (
        (3 / 2) * (fn + theta) - r * ((3 / 2) * fn + (5 / 2) * theta - (4 / 3))
    )

    f2 = (
        1
        + a * (1 + (3 / 2) * (fn + theta) - (r / 2) * (3 * fn + 5 * theta))
        + b * (3 - 4 * r)
        + (a / 2)
        * (a + 3 * b)
        * (3 - 4 * r)
        * (fn + theta - r * (fn - theta + 2 * theta**2))
    )

    f3 = 1 + a * (1 - (fn + (3 / 2) * theta) + r * (fn + theta))

    f4 = 1 + (a / 4) * (fn + 3 * theta - r * (fn - theta))

    f5 = a * (-fn + r * (fn + theta - (4 / 3))) + b * theta * (3 - 4 * r)

    f6 = 1 + a * (1 + fn - r * (fn + theta)) + b * (1 - theta) * (3 - 4 * r)

    f7 = (
        2
        + (a / 4) * (3 * fn + 9 * theta - r * (3 * fn + 5 * theta))
        + b * theta * (3 - 4 * r)
    )

    f8 = a * (1 - 2 * r + (fn / 2) * (r - 1) + (theta / 2) * (5 * r - 3)) + b * (
        1 - theta
    ) * (3 - 4 * r)

    f9 = a * ((r - 1) * fn - r * theta) + b * theta * (3 - 4 * r)

    p = f1 / f2
    q = (1 / 5) * ((2 / f3) + (1 / f4) + ((f4 * f5 + f6 * f7 - f8 * f9) / (f2 * f4)))

    return p, q



from collections.abc import Sequence
from typing import TypeVar, overload

import numpy as np
import numpy.typing as npt

from rock_physics_open.equinor_utilities.various_utilities.types import Array1D


def brie(
    s_gas: npt.NDArray[np.float64],
    k_gas: npt.NDArray[np.float64],
    s_brine: npt.NDArray[np.float64],
    k_brine: npt.NDArray[np.float64],
    s_oil: npt.NDArray[np.float64],
    k_oil: npt.NDArray[np.float64],
    e: float | npt.NDArray[np.float64],
) -> npt.NDArray[np.float64]:
    """
    Brie function for effective bulk modulus of a mix of fluids.

    Parameters
    ----------
    s_gas : np.ndarray
        Gas saturation [ratio].
    k_gas : np.ndarray
        Gas bulk modulus [Pa].
    s_brine : np.ndarray
        Brine saturation [ratio].
    k_brine : np.ndarray
        Brine bulk modulus [Pa].
    s_oil : np.ndarray
        Oil saturation [ratio].
    k_oil : np.ndarray
        Oil bulk modulus [Pa].
    e : float or np.ndarray
        Exponent in Brie function [unitless].

    Returns
    -------
    np.ndarray
        k: effective bulk modulus [Pa].
    """

    # Reuss average for fluids, catch zero fluid saturations
    idx = s_brine + s_oil > 0
    k_liquid = np.zeros(k_brine.shape)
    k_liquid[idx] = (s_brine[idx] / k_brine[idx] + s_oil[idx] / k_oil[idx]) ** -1

    return (k_liquid - k_gas) * (1 - s_gas) ** e + k_gas


def wood(
    s1: npt.NDArray[np.float64],
    k1: npt.NDArray[np.float64],
    rho1: npt.NDArray[np.float64],
    k2: npt.NDArray[np.float64],
    rho2: npt.NDArray[np.float64],
) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:
    """
    Wood effective fluid properties for a mix of two fluids.

    Parameters
    ----------
    s1 : np.ndarray
        Saturation of phase 1 [ratio].
    k1 : np.ndarray
        Bulk modulus of phase 1 [Pa].
    rho1 : np.ndarray
        Density of phase 1 [kg/m^3].
    k2 : np.ndarray
        Bulk modulus of phase 2 [Pa].
    rho2 : np.ndarray
        Density of phase 2 [kg/m^3].

    Returns
    -------
    tuple
        k, rho : np.ndarray.
        k: effective fluid bulk modulus [Pa], rho: effective fluid density [kg/m^3].
    """
    s2 = 1 - s1

    k = 1 / (s1 / k1 + s2 / k2)
    rho = s1 * rho1 + s2 * rho2

    return k, rho


_T = TypeVar("_T", float, Array1D[np.float64])


@overload
def multi_wood(
    fractions: Sequence[Array1D[np.float64]],
    bulk_moduli: Sequence[Array1D[np.float64]],
) -> Array1D[np.float64]: ...


@overload
def multi_wood(
    fractions: Sequence[float],
    bulk_moduli: Sequence[float],
) -> float: ...


def multi_wood(
    fractions: Sequence[_T],
    bulk_moduli: Sequence[_T],
) -> float | Array1D[np.float64]:
    assert len(fractions) == len(bulk_moduli)
    sum_fractions = sum(fractions)
    ratio_sum = sum(
        saturation / bulk_modulus
        for (saturation, bulk_modulus) in zip(fractions, bulk_moduli)
    )
    return sum_fractions / ratio_sum



from .array_functions import array_inverse, array_matrix_mult
from .calc_pressure import calc_pressure_vec
from .pressure_input import pressure_input_utility
from .t_matrix_vec import t_matrix_porosity_vectorised

__all__ = [
    "array_inverse",
    "array_matrix_mult",
    "calc_pressure_vec",
    "pressure_input_utility",
    "t_matrix_porosity_vectorised",
]



from typing import Literal, cast

import numpy as np
import numpy.typing as npt

from rock_physics_open.equinor_utilities import gen_utilities, std_functions


def reflectivity(
    vp_inp: npt.NDArray[np.float64],
    vs_inp: npt.NDArray[np.float64],
    rho_inp: npt.NDArray[np.float64],
    theta: float = 0.0,
    k: float = 2.0,
    model: Literal["AkiRichards", "SmithGidlow"] = "AkiRichards",
) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.bool_]]:
    """
    Reflectivity model according to Aki and Richards or Smith and Gidlow for weak contrasts
    and angles less than critical angle.

    In this function it is not allowed to have any missing values in the input logs.
    Instead of interpolating here without the user knowing, raise an input value
    exception and leave it to the user to provide complete logs.

    Parameters
    ----------
    vp_inp : np.ndarray
        Compressional wave velocity [m/s].
    vs_inp : np.ndarray
        Shear wave velocity [m/s].
    rho_inp : np.ndarray
        Bulk density [kg/m^3].
    theta : float
        Incidence angle [radians] (default value 0).
    k : float
        Background Vp/Vs ratio [ratio] (default value 2.0).
    model : str
        One of 'AkiRichards' (default) or 'SmithGidlow'.

    Returns
    -------
    tuple
        refl_coef, idx_inp : np.ndarray.
        refl_coef: reflection coefficient [ratio],
        idx_inp: index to accepted part of the input arrays [bool].
    """

    vp, vs, rho, theta_, k_ = cast(
        list[npt.NDArray[np.float64]],
        gen_utilities.dim_check_vector((vp_inp, vs_inp, rho_inp, theta, k)),
    )

    idx_inp, (vp, vs, rho, theta_, k_) = cast(
        tuple[npt.NDArray[np.bool_], list[npt.NDArray[np.float64]]],
        gen_utilities.filter_input_log(
            [vp, vs, rho, theta_, k_],
            positive=True,
        ),
    )

    if np.any(~idx_inp):
        # Only NaNs at the start or end? Find the first and last valid sample and check
        # if there are any invalid samples in between
        first_samp = np.where(idx_inp)[0][0]
        last_samp = np.where(idx_inp)[0][-1]
        if np.any(~idx_inp[first_samp : last_samp + 1]):
            # Find the culprit(s)
            idx_vp = gen_utilities.filter_input_log(
                [vp_inp[first_samp : last_samp + 1]], positive=True
            )[0]
            idx_vs = gen_utilities.filter_input_log(
                [vs_inp[first_samp : last_samp + 1]], positive=True
            )[0]
            idx_rho = gen_utilities.filter_input_log(
                [rho_inp[first_samp : last_samp + 1]], positive=True
            )[0]
            log_str = (
                int(np.any(~idx_vp)) * "Vp, "
                + int(np.any(~idx_vs)) * "Vs, "
                + int(np.any(~idx_rho)) * "Rho, "
            )
            pl_str = (
                (int(np.any(~idx_vp)) + int(np.any(~idx_vs)) + int(np.any(~idx_rho)))
                > 1
            ) * "s"
            raise ValueError(
                "{0:} reflectivity: Missing or illegal values in input log{1:}: {2:}interpolation of input log{1:} is needed\n".format(
                    model, log_str, pl_str
                )
            )

    if model == "AkiRichards":
        refl_coef = std_functions.aki_richards(vp, vs, rho, theta_, k_)
    elif model == "SmithGidlow":
        refl_coef = std_functions.smith_gidlow(vp, vs, rho, theta_, k_)
    else:
        raise ValueError(  # pyright: ignore[reportUnreachable] | Kept for backward compatibility
            f'{__file__}: unknown model: {model}, should be one of "AkiRichards", "SmithGidlow"'
        )

    return refl_coef, idx_inp



import numpy as np
import pytest
from numpy.random import default_rng

from rock_physics_open.equinor_utilities.various_utilities import reflectivity


class TestReflectivity:
    @pytest.fixture(autouse=True)
    def setup(self):
        rg = default_rng(5947037623874)
        self.vp = 3500 * (1.0 + 0.2 * rg.random(11))
        self.vs = 1200 * (1.0 + 0.4 * rg.random(11))
        self.rho = 2650 * (1.0 + 0.02 * rg.random(11))
        self.theta = 10.0
        self.k = 2.0

    def test_reflectivity_AR_ok_inputs(self):
        r_aki_rich, idx = reflectivity(
            self.vp, self.vs, self.rho, theta=self.theta, k=self.k, model="AkiRichards"
        )
        r_aki_rich_ref = np.array(
            [
                0.006805,
                0.0028961,
                0.0073632,
                -0.0016639,
                -0.0053192,
                -0.0114692,
                0.0059247,
                -0.0300648,
                0.015451,
                0.0143082,
                0.0143082,
            ]
        )
        idx_ref = np.ones(11).astype(bool)
        np.testing.assert_almost_equal(r_aki_rich, r_aki_rich_ref)
        np.testing.assert_almost_equal(idx, idx_ref)

    def test_reflectivity_SG_ok_inputs(self):
        r_sm_gi, idx = reflectivity(
            self.vp, self.vs, self.rho, theta=self.theta, k=self.k, model="SmithGidlow"
        )
        r_sm_gi_ref = np.array(
            [
                0.0041067,
                0.0033311,
                0.012066,
                -0.0041207,
                -0.0024095,
                -0.0165878,
                0.0082089,
                -0.0389886,
                0.0207101,
                0.0145037,
                0.0145037,
            ]
        )
        idx_ref = np.ones(11).astype(bool)
        np.testing.assert_almost_equal(r_sm_gi, r_sm_gi_ref)
        np.testing.assert_almost_equal(idx, idx_ref)

    def test_reflectivity_nan(self):
        self.vp[5] = np.nan
        with pytest.raises(ValueError, match="Missing or illegal values in input"):
            _ = reflectivity(
                vp_inp=self.vp,
                vs_inp=self.vs,
                rho_inp=self.rho,
                theta=self.theta,
                k=self.k,
                model="AkiRichards",
            )



from .cemented_shalysand_sandyshale_models import cemented_shaly_sand_sandy_shale_model
from .constant_cement_models import (
    constant_cement_model,
    constant_cement_model_dry,
)
from .constant_cement_optimisation import constant_cement_model_optimisation
from .contact_cement_model import contact_cement_model
from .curvefit_sandstone_models import (
    curvefit_constant_cement,
    curvefit_friable,
    curvefit_patchy_cement,
)
from .friable_models import (
    friable_model,
    friable_model_dry,
)
from .friable_optimisation import friable_model_optimisation
from .friable_shalysand_sandyshale_models import friable_shaly_sand_sandy_shale_model
from .patchy_cement_fluid_substitution_model import (
    patchy_cement_pressure_fluid_substitution,
)
from .patchy_cement_model import (
    constant_cement_model_pcm,
    patchy_cement_model_cem_frac,
    patchy_cement_model_dry,
    patchy_cement_model_weight,
)
from .patchy_cement_optimisation import (
    patchy_cement_model_optimisation,
    patchy_cement_model_optimisation_multiwell,
)
from .unresolved_cemented_sandshale_models import (
    unresolved_constant_cement_sand_shale_model,
)
from .unresolved_friable_sandshale_models import unresolved_friable_sand_shale_model

__all__ = [
    "cemented_shaly_sand_sandy_shale_model",
    "constant_cement_model",
    "constant_cement_model_dry",
    "constant_cement_model_optimisation",
    "contact_cement_model",
    "curvefit_constant_cement",
    "curvefit_friable",
    "curvefit_patchy_cement",
    "friable_model",
    "friable_model_dry",
    "friable_model_optimisation",
    "friable_shaly_sand_sandy_shale_model",
    "patchy_cement_pressure_fluid_substitution",
    "constant_cement_model_pcm",
    "patchy_cement_model_dry",
    "patchy_cement_model_cem_frac",
    "patchy_cement_model_weight",
    "patchy_cement_model_optimisation",
    "patchy_cement_model_optimisation_multiwell",
    "unresolved_constant_cement_sand_shale_model",
    "unresolved_friable_sand_shale_model",
]



import numpy as np
import numpy.typing as npt


def dead_oil_velocity(
    temperature: npt.NDArray[np.float64],
    pressure: npt.NDArray[np.float64],
    reference_density: npt.NDArray[np.float64],
) -> npt.NDArray[np.float64]:
    """
    The primary wave velocity in oil without dissolved gas (dead).

    Uses equation 20a from Batzle & Wang [1].

    :param reference_density: Density of oil at 15.6 degrees Celsius and atmospheric
        pressure [kg/m^3]
    :param pressure: Formation pressure [Pa] of oil
    :param temperature: Temperature [°C] of oil.
    :return: primary velocity of dead oil in m/s.
    """
    pressure_mpa = pressure * 1e-6
    density_gcc = reference_density / 1000.0
    return (
        2096 * np.sqrt(density_gcc / (2.6 - density_gcc))
        - 3.7 * temperature
        + 4.64 * pressure_mpa
        + 0.0115
        * (4.12 * np.sqrt(1.08 * density_gcc**-1 - 1) - 1)
        * temperature
        * pressure_mpa
    )



from typing import cast

import numpy as np
import numpy.typing as npt
from matplotlib import pyplot as plt
from matplotlib.colorbar import Colorbar
from matplotlib.text import Text

from .ternary_plot_utilities import set_ternary_figure, triangle_transform


def shale_prop_ternary(
    quartz: npt.NDArray[np.float64],
    carb: npt.NDArray[np.float64],
    clay: npt.NDArray[np.float64],
    kero: npt.NDArray[np.float64],
    phit: npt.NDArray[np.float64],
    col_code: npt.NDArray[np.float64],
    name_col_code: str,
    well_name: str,
    draw_figures: bool = True,
) -> npt.NDArray[np.float64]:
    """Calculate hardness.

    Parameters
    ----------
    quartz : np.ndarray
        Quartz volume fraction [fraction].
    carb : np.ndarray
        Carbonate volume fraction [fraction].
    clay : np.ndarray
        Clay volume fraction [fraction].
    kero : np.ndarray
        Kerogen volume fraction [fraction].
    phit : np.ndarray
        Porosity [fraction].
    col_code : np.ndarray
        Property used for colour coding [unknown].
    name_col_code : str
        Plot annotation of log used for colour coding.
    well_name : str
        Plot heading with well name.
    draw_figures : bool
        Decide if figures are drawn or not, default is True.

    Returns
    -------
    np.ndarray
        hardness [float].
    """
    KEROMAX = 0.25

    _, _ = set_ternary_figure(
        delta_x=1,
        delta_y=1,
        title="Hardness vs. Organic Content Ternary Plot",
        well_name=well_name,
    )

    # Define patches
    col = (
        1
        / 255
        * np.array(
            [
                [
                    208,
                    208,
                    208,
                ],
                [
                    198,
                    198,
                    198,
                ],
                [
                    189,
                    189,
                    189,
                ],
                [
                    179,
                    179,
                    179,
                ],
                [
                    169,
                    169,
                    169,
                ],
                [
                    223,
                    213,
                    203,
                ],
                [
                    214,
                    203,
                    193,
                ],
                [
                    204,
                    194,
                    183,
                ],
                [
                    195,
                    183,
                    174,
                ],
                [
                    187,
                    172,
                    164,
                ],
                [
                    239,
                    218,
                    197,
                ],
                [
                    229,
                    208,
                    188,
                ],
                [
                    221,
                    197,
                    178,
                ],
                [
                    213,
                    186,
                    169,
                ],
                [
                    205,
                    174,
                    159,
                ],
                [
                    254,
                    223,
                    192,
                ],
                [
                    246,
                    211,
                    182,
                ],
                [
                    239,
                    200,
                    173,
                ],
                [
                    231,
                    189,
                    164,
                ],
                [223, 177, 154],
            ]
        )
    )

    corner_points = np.array(
        [
            [0.00, 0.00],
            [0.25, 0.00],
            [0.25, 0.20],
            [0.00, 0.20],
        ]
    )

    start_x, start_y = np.meshgrid(
        np.arange(0, 1, 0.25), np.arange(0, 1, 0.2), indexing="ij"
    )
    start_x = start_x.flatten()
    start_y = start_y.flatten()
    vertices = [corner_points + [start_x[i], start_y[i]] for i in range(len(start_x))]
    vertices_xy = [triangle_transform(v) for v in vertices]

    _ = [
        plt.fill(vertices_xy[i][:, 0], vertices_xy[i][:, 1], color=col[i, :])
        for i in range(len(vertices))
    ]

    # Draw lines around patches
    x = np.array([[0], [1]], dtype=np.float64)
    y = np.array([[0], [0]], dtype=np.float64)
    for i in np.arange(0, 1, 0.2):
        xy = np.column_stack((x, y + i))
        t_xy = triangle_transform(xy)
        _ = plt.plot(t_xy[:, 0], t_xy[:, 1], "k", linewidth=1)
    x = np.array([[0], [0]], dtype=np.float64)
    y = np.array([[0], [1]], dtype=np.float64)
    for i in np.arange(0, 1.1, 0.25):
        xy = np.column_stack((x + i, y))
        t_xy = triangle_transform(xy)
        _ = plt.plot(t_xy[:, 0], t_xy[:, 1], "k", linewidth=1)

    # Annotate axes
    xi = np.linspace(0, 1, 5)
    yi = np.linspace(0, 0.25, 6)
    for j in range(len(xi)):
        _ = plt.text(xi[j], -0.025, "%.2f" % (xi[j]))
    for j in range(len(yi)):
        _ = plt.text(
            1 - 0.5 / 0.25 * yi[j] + 0.015,
            yi[j] * np.sqrt(3) / 2 / 0.25,
            "%.2f" % (yi[j]),
        )
    handles: list[Text] = []
    # Axes labels
    handles.append(
        plt.text(
            x=0.5,
            y=-0.07,
            s="Hardness percentage\nClay+TOC+Phit <== ==> Quartz+Feldspar+Carbonate",
            horizontalalignment="center",
            verticalalignment="center",
        )
    )
    handles.append(
        plt.text(
            x=0.95,
            y=np.sqrt(3) / 4,
            s="Organic percentage",
            horizontalalignment="center",
            verticalalignment="center",
            rotation=-62,
        )
    )
    plt.setp(handles, fontname="sans-serif", fontweight="bold", fontsize=11)

    # Region boundaries
    df = 0.1 * np.array([1, np.tan(np.pi / 6)])
    bb = np.array(
        [
            [1.00, 0.20],
            [1.00, 0.60],
        ]
    )
    bbt = triangle_transform(bb)
    bet = bbt + df
    for j in range(bb.shape[0]):
        _ = plt.plot(
            [
                bbt[j, 0],
                bet[j, 0],
            ],
            [
                bbt[j, 1],
                bet[j, 1],
            ],
            "k",
            linewidth=0.5,
        )
    df = 0.1 * np.array([-1, np.tan(np.pi / 6)])
    bb = np.array(
        [
            [0.00, 0.20],
            [0.00, 0.60],
        ]
    )
    bbt = triangle_transform(bb)
    bet = bbt + df
    for j in range(bb.shape[0]):
        _ = plt.plot(
            [
                bbt[j, 0],
                bet[j, 0],
            ],
            [
                bbt[j, 1],
                bet[j, 1],
            ],
            "k",
            linewidth=0.5,
        )

    # Annotate regions
    names = [
        "Organically lean\nclaystones",
        "Organically rich\nclaystones",
        "Organically dominated\nclaystones",
        "Organically dominated\nmudstones",
        "Organically rich\nmudstones",
        "Organically lean\nmudstones",
    ]
    text_ang = [60, 60, 60, -60, -60, -60]
    text_pos = np.array(
        [
            [0.0, 0.1],
            [0.0, 0.4],
            [0.0, 0.8],
            [1.0, 0.8],
            [1.0, 0.4],
            [1.0, 0.1],
        ]
    )
    text_pos = triangle_transform(text_pos)
    text_pos = text_pos + np.array(
        [
            [-0.05, 0.0],
            [0.1, 0.05],
        ]
    ).reshape(2, 2).repeat(3, axis=0)
    for i in range(len(names)):
        _ = plt.text(
            x=text_pos[i, 0],
            y=text_pos[i, 1],
            s=names[i],
            horizontalalignment="center",
            verticalalignment="center",
            fontsize=10,
            rotation=text_ang[i],
        )

    # Calculate the hardness parameter
    soft = clay + kero + phit
    brit = quartz + carb

    hard = brit / (brit + soft)

    # Scale kerogen
    kero = np.minimum(kero, KEROMAX)
    kero_scal = kero / KEROMAX

    data_xy = triangle_transform(np.column_stack((hard, kero_scal)))
    _ = plt.scatter(data_xy[:, 0], data_xy[:, 1], s=64, c=col_code, zorder=4)

    hcb = cast(  # Casting due to incomplete typing in matplotlib
        Colorbar,
        plt.colorbar(pad=0.1),
    )
    hcb.set_label(
        label=name_col_code,
        fontname="sans-serif",
        fontweight="bold",
        fontsize=11,
    )

    # plt.show()<= Show command is issued in controlling class method
    if draw_figures:
        plt.draw()

    return hard



import numpy as np

from rock_physics_open.equinor_utilities.various_utilities.types import Array1D, Array3D

from .array_functions import array_inverse, array_matrix_mult
from .g_tensor import g_tensor_vec


def pressure_input_utility(
    k_min: Array1D[np.float64],
    mu_min: Array1D[np.float64],
    log_length: int,
) -> tuple[Array3D[np.float64], Array3D[np.float64], Array3D[np.float64]]:
    """Utility that calculates some of the elastic properties needed for T-Matrix pressure estimation.

    Parameters
    ----------
    k_min : np.ndarray
        Effective mineral bulk modulus [Pa].
    mu_min : np.ndarray
        Effective mineral shear modulus [Pa].
    log_length : int
        Number of samples in logs.

    Returns
    -------
    tuple
        c0 (background stiffness), s0 (inverse of background stiffness), gd (G tensor for inclusions with aspect
        ratio 1.0).
    """
    # Calculate elastic parameters
    c11 = k_min + 4 / 3 * mu_min
    c44 = mu_min
    c12 = c11 - 2 * c44

    i4 = np.tile(np.eye(6).reshape(1, 6, 6), (log_length, 1, 1))
    c0 = np.zeros((log_length, 6, 6))

    for i in range(3):
        c0[:, i, i] = c11
        c0[:, i + 3, i + 3] = 2 * c44
    for i in range(2):
        c0[:, 0, i + 1] = c12
        c0[:, 1, 2 * i] = c12
        c0[:, 2, i] = c12

    s0 = array_matrix_mult(i4, array_inverse(c0))
    gd = g_tensor_vec(c0=c0, s_0=s0, alpha=1.0)

    return c0, s0, gd



import os

import pytest
from numpy.random import default_rng

from rock_physics_open.equinor_utilities.snapshot_test_utilities import (
    INITIATE,
    compare_snapshots,
    get_snapshot_name,
    read_snapshot,
    store_snapshot,
)
from rock_physics_open.ternary_plots import run_ternary

rg = default_rng(5947037623874)
quartz = rg.random(11)
carb = rg.random(11)
clay = rg.random(11)
kero = rg.random(11)
phi = rg.random(11)
misc = rg.random(11)
misc_log_type = "Vp"
well_name = "35_11_15"


@pytest.mark.use_graphics
def test_ternary():
    args = run_ternary(
        quartz,
        carb,
        clay,
        kero,
        phi,
        misc,
        misc_log_type,
        well_name,
        draw_figures=False,
    )

    if not os.path.isfile(get_snapshot_name()) or INITIATE:
        _ = store_snapshot(get_snapshot_name(), *args)
    else:
        assert compare_snapshots(args, read_snapshot(get_snapshot_name()))



import os
from pathlib import Path

import numpy as np
import pandas as pd

from rock_physics_open.equinor_utilities.snapshot_test_utilities import (
    INITIATE,
    compare_snapshots,
    get_snapshot_name,
    read_snapshot,
    store_snapshot,
)
from rock_physics_open.sandstone_models import (
    constant_cement_model_optimisation,
    friable_model_optimisation,
    patchy_cement_model_optimisation,
)

from ..conftest import TESTDATA_DIR

data_df = pd.read_csv(TESTDATA_DIR / "sandstone_optimisation.csv")

vp = data_df["VP"].to_numpy()
vs = data_df["VS"].to_numpy()
rhob = data_df["RHOB"].to_numpy()
phit = data_df["PHIT"].to_numpy()

k_min = 36.8e9 * np.ones_like(phit)
mu_min = 44.0e9 * np.ones_like(phit)
rho_min = 2650 * np.ones_like(phit)
k_cem = 36.8e9 * np.ones_like(phit)
mu_cem = 44.0e9 * np.ones_like(phit)
rho_cem = 2650 * np.ones_like(phit)
k_fl = 2.7e9 * np.ones_like(phit)
rho_fl = 1005 * np.ones_like(phit)
p_eff = 20.0e6 * np.ones_like(phit)
phi_c = 0.45


def test_friable_optimisation(data_dir: Path):
    file_name = str(data_dir.joinpath("friable_model_optimisation.pkl"))

    args = friable_model_optimisation(
        k_min=k_min,
        mu_min=mu_min,
        rho_min=rho_min,
        k_fl=k_fl,
        rho_fl=rho_fl,
        por=phit,
        p_eff=p_eff,
        vp=vp,
        vs=vs,
        rhob=rhob,
        file_out_str=file_name,
    )

    if not os.path.isfile(get_snapshot_name()) or INITIATE:
        _ = store_snapshot(get_snapshot_name(), *args)
    else:
        assert compare_snapshots(args, read_snapshot(get_snapshot_name()))


def test_constant_cement_optimisation(data_dir: Path):
    file_name = str(data_dir.joinpath("constant_cement_model_optimisation.pkl"))

    args = constant_cement_model_optimisation(
        k_min=k_min,
        mu_min=mu_min,
        rho_min=rho_min,
        k_cem=k_cem,
        mu_cem=mu_cem,
        rho_cem=rho_cem,
        k_fl=k_fl,
        rho_fl=rho_fl,
        por=phit,
        vp=vp,
        vs=vs,
        rhob=rhob,
        file_out_str=file_name,
    )

    if not os.path.isfile(get_snapshot_name()) or INITIATE:
        _ = store_snapshot(get_snapshot_name(), *args)
    else:
        assert compare_snapshots(args, read_snapshot(get_snapshot_name()))


def test_patchy_cement_optimisation(data_dir: Path):
    file_name = str(data_dir.joinpath("patchy_cement_model_optimisation.pkl"))

    args = patchy_cement_model_optimisation(
        k_min=k_min,
        mu_min=mu_min,
        rho_min=rho_min,
        k_cem=k_cem,
        mu_cem=mu_cem,
        rho_cem=rho_cem,
        k_fl=k_fl,
        rho_fl=rho_fl,
        por=phit,
        p_eff=p_eff,
        vp=vp,
        vs=vs,
        rhob=rhob,
        phi_c=phi_c,
        file_out_str=file_name,
    )

    if not os.path.isfile(get_snapshot_name()) or INITIATE:
        _ = store_snapshot(get_snapshot_name(), *args)
    else:
        assert compare_snapshots(args, read_snapshot(get_snapshot_name()))



import numpy as np

from rock_physics_open.equinor_utilities.various_utilities.types import Array1D, Array3D


def velocity_vti_angles_vec(
    c_eff: Array3D[np.float64],
    rho_eff: Array1D[np.float64],
    angle: float,
) -> tuple[Array1D[np.float64], Array3D[np.float64], Array3D[np.float64]]:
    """Returns the P-velocity and  S-velocities.

    Parameters
    ----------
    c_eff : np.ndarray
        Effective stiffness tensor (nx6x6 matrix).
    rho_eff : np.ndarray
        Effective density.
    angle : float
        The angle between the wave vector and the axis of symmetry.

    Returns
    -------
    tuple[np.ndarray, np.ndarray, np.ndarray]
        vp_out, vsv_out, vsh_out
        vp_out, vsv_out, vsh_out : p-velocity, vertical polarisation s-velocity, horizontal polarisation s-velocity.
    """
    if not (
        c_eff.ndim == 3
        and rho_eff.ndim == 1
        and isinstance(angle, (float, int))  # pyright: ignore[reportUnnecessaryIsInstance] | kept for backwards compatibility
        and (c_eff.shape[0] == rho_eff.shape[0] and c_eff.shape[1] == c_eff.shape[2])
    ):
        raise ValueError("velocity_vti_angles: inconsistencies in input shapes")

    # vp45
    m45 = (
        (c_eff[:, 0, 0] - c_eff[:, 3, 3] / 2.0) * 0.5
        - (c_eff[:, 2, 2] - c_eff[:, 3, 3] / 2.0) * 0.5
    ) ** 2 + (c_eff[:, 0, 0] - c_eff[:, 3, 3] / 2.0) ** 2
    m45 = np.sqrt(m45)
    v2p = ((c_eff[:, 0, 0] + c_eff[:, 2, 2] + c_eff[:, 3, 3]) / 2.0 + m45) / (
        2.0 * rho_eff
    )
    sp = np.real(1.0 / (np.sqrt(v2p)))
    vp45 = 1.0 / sp

    # c11 = rho_eff * Vp90 ** 2
    c11 = np.real(c_eff[:, 0, 0])
    # c33 = rho_eff * Vp0 ** 2
    c33 = np.real(c_eff[:, 2, 2])
    # c44 = rho_eff * Vs0 ** 2
    c44 = np.real(c_eff[:, 3, 3] / 2.0)
    # c66 = rho_eff * Vs90 ** 2
    c66 = np.real(c_eff[:, 5, 5] / 2.0)

    c13 = -c44 + np.sqrt(
        4 * (rho_eff**2) * (vp45**4)
        - 2.0 * rho_eff * (vp45**2) * (c11 + c33 + 2.0 * c44)
        + (c11 + c44) * (c33 + c44)
    )

    rad_angle = (angle * np.pi) / 180.0

    m_real = (
        (c11 - c44) * (np.sin(rad_angle) ** 2) - (c33 - c44) * (np.cos(rad_angle) ** 2)
    ) ** 2 + ((c13 + c44) ** 2) * (np.sin(2.0 * rad_angle) ** 2)

    vp_out = np.sqrt(
        c11 * (np.sin(rad_angle) ** 2)
        + c33 * (np.cos(rad_angle) ** 2)
        + c44
        + np.sqrt(m_real)
    ) * np.sqrt(1.0 / (2.0 * rho_eff))

    vsv_out = np.sqrt(
        c11 * (np.sin(rad_angle) ** 2)
        + c33 * (np.cos(rad_angle) ** 2)
        + c44
        - np.sqrt(m_real)
    ) * np.sqrt(1.0 / (2.0 * rho_eff))

    vsh_out = np.sqrt(
        (c66 * (np.sin(rad_angle) ** 2) + c44 * (np.cos(rad_angle) ** 2)) / rho_eff
    )

    return vp_out, vsv_out, vsh_out



import numpy as np
import numpy.typing as npt


def voigt(
    k1: npt.NDArray[np.float64],
    mu1: npt.NDArray[np.float64],
    k2: npt.NDArray[np.float64],
    mu2: npt.NDArray[np.float64],
    f1: npt.NDArray[np.float64],
) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:
    """
    Effective media calculation of a mixture of two phases using the Voigt bound.

    Parameters
    ----------
    k1 : np.ndarray
        Bulk modulus of phase 1 [Pa].
    mu1 : np.ndarray
        Shear modulus of phase 1 [Pa].
    k2 : np.ndarray
        Bulk modulus of phase 2 [Pa].
    mu2 : np.ndarray
        Shear modulus of phase 2 [Pa].
    f1 : np.ndarray
        Fraction of phase 1 [fraction].

    Returns
    -------
    tuple
        k, mu : np.ndarray.
        k: effective bulk modulus [Pa], mu: effective shear modulus [Pa]
    """
    f2 = 1 - f1
    k_v = f1 * k1 + f2 * k2
    mu_v = f1 * mu1 + f2 * mu2

    return k_v, mu_v


def voigt_reuss_hill(
    k1: npt.NDArray[np.float64],
    mu1: npt.NDArray[np.float64],
    k2: npt.NDArray[np.float64],
    mu2: npt.NDArray[np.float64],
    f1: npt.NDArray[np.float64],
) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:
    """
    Effective media calculation of a mixture of two phases using the Voigt-Reuss-Hill average.

    Parameters
    ----------
    k1 : np.ndarray
        Bulk modulus of phase 1 [Pa].
    mu1 : np.ndarray
        Shear modulus of phase 1 [Pa].
    k2 : np.ndarray
        Bulk modulus of phase 2 [Pa].
    mu2 : np.ndarray
        Shear modulus of phase 2 [Pa].
    f1 : np.ndarray
        Fraction of phase 1 [fraction].

    Returns
    -------
    tuple
        k, mu : np.ndarray
        k: effective bulk modulus [Pa], mu: effective shear modulus [Pa]
    """
    k_v, mu_v = voigt(k1, mu1, k2, mu2, f1)
    k_r, mu_r = reuss(k1, mu1, k2, mu2, f1)

    k_vrh = (k_r + k_v) / 2
    mu_vrh = (mu_r + mu_v) / 2

    return k_vrh, mu_vrh


def multi_voigt_reuss_hill(
    *varargin: npt.NDArray[np.float64],
) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:
    """
    Voigt-Reuss-Hill with multiple mineral input.

    Call function with [k, mu] = multi_voigt_reuss_hill(k1, mu1, f2, ....kn,
    mun, fn). Fractions must add up to 1.0. All inputs are assumed to be
    vectors with the same length.

    Parameters
    ----------
    varargin : np.ndarray

    Returns
    -------
    tuple
        k, mu : np.ndarray
        k: effective bulk modulus [Pa], mu: effective shear modulus [Pa].
    """
    k_min = np.array(varargin[::3])
    mu_min = np.array(varargin[1::3])
    f = np.array(varargin[2::3])

    k_voigt = np.sum(k_min * f, axis=0)
    mu_voigt = np.sum(mu_min * f, axis=0)

    k_reuss = 1 / np.sum(f / k_min, axis=0)
    mu_reuss = 1 / np.sum(f / mu_min, axis=0)

    k = (k_voigt + k_reuss) / 2
    mu = (mu_voigt + mu_reuss) / 2

    return k, mu


def reuss(
    k1: npt.NDArray[np.float64],
    mu1: npt.NDArray[np.float64],
    k2: npt.NDArray[np.float64],
    mu2: npt.NDArray[np.float64],
    f1: npt.NDArray[np.float64],
) -> tuple[npt.NDArray[np.float64], npt.NDArray[np.float64]]:
    """
    Effective media calculation of a mixture of two phases using the Reuss bound.

    Parameters
    ----------
    k1 : np.ndarray
        Bulk modulus of phase 1 [Pa].
    mu1 : np.ndarray
        Shear modulus of phase 1 [Pa].
    k2 : np.ndarray
        Bulk modulus of phase 2 [Pa].
    mu2 : np.ndarray
        Shear modulus of phase 2 [Pa].
    f1 : np.ndarray
        Fraction of phase 1 [fraction].

    Returns
    -------
    tuple
        k, mu : np.ndarray.
        k: effective bulk modulus [Pa], mu: effective shear modulus [Pa].
    """
    f2 = 1 - f1

    k_r = 1 / (f1 / k1 + f2 / k2)
    mu_r = 1 / (f1 / mu1 + f2 / mu2)

    return k_r, mu_r



import numpy as np
import numpy.typing as npt


def pressure_adjusted_dead_oil_density(
    pressure: npt.NDArray[np.float64],
    reference_density: npt.NDArray[np.float64],
) -> npt.NDArray[np.float64]:
    """
    Adjusts density of a dead oil (without dissolved gas) to a given pressure.

    Uses equation 18 from Batzle & Wang [1].

    :param reference_density: The density [kg/m^3] of the dead oil at 15.6 degrees Celsius
        and atmospheric pressure.
    :param pressure: Formation pressure [Pa] to adjust to.
    :return: Density of oil at given pressure and 21 degrees Celsius (~70 degrees
    Fahrenheit). [kg/m^3]
    """
    pressure_mpa = pressure / 1e6
    density_gcc = reference_density / 1000.0
    return 1000.0 * (
        density_gcc
        + (0.00277 * pressure_mpa - 1.71e-7 * pressure_mpa**3)
        * (density_gcc - 1.15) ** 2
        + 3.49e-4 * pressure_mpa
    )


def temperature_adjusted_dead_oil_density(
    temperature: npt.NDArray[np.float64],
    density_at_21c: npt.NDArray[np.float64],
) -> npt.NDArray[np.float64]:
    """
    Adjusts density of a dead oil (without dissolved gas) to a given temperature.

    Uses equation 19 from Batzle & Wang [1].

    :param density_at_21c: The density [kg/m^3] of the dead oil at 21 degrees Celsius
    :param temperature: Temperature [°C] of oil.
    :return: Density of oil at given temperature. [kg/m^3]
    """
    density_at_21c_gcc = density_at_21c / 1000.0
    return (
        1000.0 * density_at_21c_gcc / (0.972 + 3.81e-4 * (temperature + 17.78) ** 1.175)
    )


def dead_oil_density(
    temperature: npt.NDArray[np.float64],
    pressure: npt.NDArray[np.float64],
    reference_density: npt.NDArray[np.float64],
) -> npt.NDArray[np.float64]:
    """
    The density of oil without dissolved gas (dead).

    Uses equation 18 & 19 from Batzle & Wang [1].

    :param reference_density: Density of oil at 15.6 degrees Celsius and atmospheric
        pressure [kg/m^3]
    :param pressure: Formation pressure [Pa] of oil
    :param temperature: Temperature [°C] of oil.
    :return: density of dead oil at given conditions (kg/m^3).
    """
    density_p = pressure_adjusted_dead_oil_density(
        pressure=pressure,
        reference_density=reference_density,
    )
    return temperature_adjusted_dead_oil_density(
        temperature=temperature,
        density_at_21c=density_p,
    )



from typing import Any

import numpy as np
import numpy.typing as npt
import pandas as pd


def filter_output(
    idx_inp: npt.NDArray[np.bool_],
    inp_log: list[npt.NDArray[Any] | pd.DataFrame]
    | tuple[npt.NDArray[Any] | pd.DataFrame, ...]
    | npt.NDArray[Any]
    | pd.DataFrame,
) -> list[npt.NDArray[Any] | pd.DataFrame]:
    """
    Function to restore outputs from a plugin to original length and
    with values at correct positions. The logs are assumed to go through
    matching input filtering done by gen_utilities.filter_input_log earlier.

    Parameters
    ----------
    idx_inp: np.ndarray
        boolean array which is True at locations to be filled, length idx_inp is returned length of
        arrays or data frames.
    inp_log: tuple or list or np.ndarray or pd.DataFrame
        input numpy array(s) or pandas data frame(s), in list or tuple that are to be expanded to original
        length.

    Returns
    -------
    return_logs : list
        Expanded inputs.
    """

    def _expand_array(
        idx: npt.NDArray[np.bool_], inp_single_log: npt.NDArray[Any]
    ) -> npt.NDArray[Any]:
        logs = np.ones(idx.shape, dtype=float) * np.nan
        try:
            logs[idx] = inp_single_log.flatten()
        except ValueError:
            # Assume that the dtype  of the input log is not fit for casting to float, set to object and retry
            logs = logs.astype(object)
            logs[idx] = inp_single_log
        return logs.reshape(idx.shape)

    def _expand_df(idx: npt.NDArray[np.bool_], inp_df: pd.DataFrame) -> pd.DataFrame:
        logs = pd.DataFrame(
            columns=inp_df.columns, index=np.arange(idx.shape[0], dtype=np.intp)
        )
        logs.loc[idx] = inp_df
        return logs

    if not isinstance(inp_log, (list, tuple, np.ndarray, pd.DataFrame)):  # pyright: ignore[reportUnnecessaryIsInstance] | Kept for backward compatibility
        raise ValueError(  # pyright: ignore[reportUnreachable] | Kept for backward compatibility
            "filter_output: unknown input data type: {}".format(type(inp_log))
        )
    if not isinstance(idx_inp, (list, np.ndarray)):  # pyright: ignore[reportUnnecessaryIsInstance] | Kept for backward compatibility
        raise ValueError(  # pyright: ignore[reportUnreachable] | Kept for backward compatibility
            "filter_output: unknown filter array data type: {}".format(type(idx_inp))
        )

    # Make iterable in case of single input
    if isinstance(inp_log, (np.ndarray, pd.DataFrame)):
        inp_log = [inp_log]

    if isinstance(idx_inp, np.ndarray):  # pyright: ignore[reportUnnecessaryIsInstance] | Kept for backward compatibility
        idx_inp_ = [idx_inp]

    # Possible to simplify?
    if len(idx_inp_) != len(inp_log):
        if len(idx_inp_) == 1:
            idx_inp_ = idx_inp_ * len(inp_log)
        else:
            raise ValueError(
                "filter_output: mismatch between length of filter arrays and inputs: {} and {}".format(
                    len(idx_inp), len(inp_log)
                )
            )

    return_logs: list[npt.NDArray[Any] | pd.DataFrame] = []
    for this_idx, this_log in zip(idx_inp_, inp_log):
        if isinstance(this_log, np.ndarray):
            return_logs.append(_expand_array(this_idx, this_log))
        elif isinstance(this_log, pd.DataFrame):  # pyright: ignore[reportUnnecessaryIsInstance] | Kept for backward compatibility
            return_logs.append(_expand_df(this_idx, this_log))

    return return_logs



import numpy as np
import numpy.typing as npt

from .dead_oil_velocity import dead_oil_velocity
from .live_oil_density import live_oil_pseudo_density


def live_oil_velocity(
    temperature: npt.NDArray[np.float64],
    pressure: npt.NDArray[np.float64],
    reference_density: npt.NDArray[np.float64],
    gas_oil_ratio: npt.NDArray[np.float64],
    gas_gravity: npt.NDArray[np.float64],
) -> npt.NDArray[np.float64]:
    """
    Primary wave velocity of live oil at saturation.

    Substitute Equation 22 in Equation 20 of Batzle & Wang [1].

    :param reference_density: Density of the oil without dissolved gas
        at 15.6 degrees Celsius and atmospheric pressure. [kg/m^3]
    :param pressure: Formation pressure [Pa] of oil
    :param gas_oil_ratio: The volume ratio of gas to oil [l/l]
    :param temperature: Temperature [°C] of oil.
    :param gas_gravity: molar mass of gas relative to air molar mas.
    :return: Primary wave velocity of live oil [m/s].
    """
    rho_marked = live_oil_pseudo_density(
        temperature=temperature,
        reference_density=reference_density,
        gas_oil_ratio=gas_oil_ratio,
        gas_gravity=gas_gravity,
    )
    return dead_oil_velocity(
        temperature=temperature,
        pressure=pressure,
        reference_density=rho_marked,
    )



import ctypes
import os
import platform
import re
from typing import Any

import numpy as np
import numpy.typing as npt
from matplotlib import pyplot as plt
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from matplotlib.lines import Line2D
from matplotlib.text import Text

from rock_physics_open.equinor_utilities.various_utilities.types import Array1D, Array2D

HORZRES = 8
VERTRES = 10

# The following code is only needed if the figure size is given in inches
# LOGPIXELSX = 88
# LOGPIXELSY = 90


def set_ternary_figure(
    delta_x: int,
    delta_y: int,
    title: str,
    well_name: str,
) -> tuple[Figure, Axes]:
    """Find screen size, and make a suitable figure size and position, shift window with
    number of steps given by deltaX and deltaY.
    """

    if platform.system() == "Windows":
        dc = ctypes.windll.user32.GetDC(0)
        pix_x = ctypes.windll.gdi32.GetDeviceCaps(dc, HORZRES)
        pix_y = ctypes.windll.gdi32.GetDeviceCaps(dc, VERTRES)
    elif platform.system() == "Linux":
        f = os.popen("xrandr | grep '*'")
        pix_x, pix_y = np.array(re.findall(r"\d+", f.read()), dtype=int)[0:2]
    elif platform.system() == "Darwin":
        f = os.popen("system_profiler SPDisplaysDataType | grep Resolution")
        pix_x, pix_y = np.array(re.findall(r"\d+", f.read()), dtype=int)[0:2]
    else:
        raise ValueError("Unrecognised operating system")

    # Add figure
    fig = plt.figure(title, facecolor=(0.9, 0.9, 0.9))
    # Set figure size and position
    mngr = plt.get_current_fig_manager()
    if platform.system() == "Windows":
        wid = int(pix_x * 0.4)
        hei = int(pix_y * 0.6)
        start_x = int(pix_x * (0.1 + delta_x / 5))
        start_y = int(pix_y * (0.1 + delta_y / 10))
        mngr.window.wm_geometry(  # pyright: ignore[reportAttributeAccessIssue, reportUnknownMemberType] | Incomplete type hints in matplotlib
            "{}x{}+{}+{}".format(wid, hei, start_x, start_y),
        )
    else:
        wid = int(pix_x * 0.2)
        hei = int(pix_y * 0.6)
        start_x = int(pix_x * (0.1 + delta_x / 5))
        start_y = int(pix_y * (0.1 + delta_y / 10))
        mngr.window.wm_geometry(  # pyright: ignore[reportAttributeAccessIssue, reportUnknownMemberType] | Incomplete type hints in matplotlib
            "{}x{}+{}+{}".format(wid, hei, start_x, start_y),
        )

    # Set default colour map
    plt.jet()

    ax = plt.axes(frameon=False, aspect="equal")
    _ = ax.set_title(well_name)
    _ = plt.ylim((-0.02, 1))
    _ = plt.xlim((0, 1))
    ax.set_axis_off()

    return fig, ax


def ternary_coord_trans(
    *args: Array1D[np.float64] | Array2D[np.float64],
) -> Array2D[np.float64]:
    """Routine to transform ternary coordinates to xy coordinates.
    Inputs can either be 3 separate coordinate arrays or a nX3 array
    The sum of the input coordinates should be one - the routine will normalise the inputs

    Returns
    -------
    np.ndarray
        Coordinates in nx2 numpy array.
    """
    trans_mat = np.array(
        [
            [0.0, 0.0],
            [1.0, 0.0],
            [0.5, 0.5 * np.sqrt(3)],
        ]
    )

    tern_coord = None
    if len(args) == 3:
        try:
            tern_coord = np.array(np.column_stack(args[:]))
        except ValueError:
            raise ValueError(
                f"{__file__}: Could not combine inputs to _ternary_coord_trans routine"
            )
    elif len(args) == 1:
        tern_coord = np.array(args[0])
    else:
        raise ValueError(f"{__file__}: Unexpected input to TernaryCoordTrans routine")

    # Normalise inputs
    tot = tern_coord.sum(axis=1).reshape(tern_coord.shape[0], 1)
    tern_coord = tern_coord / tot

    return np.array(tern_coord @ trans_mat)


def triangle_transform(xy1: npt.NDArray[np.float64]) -> npt.NDArray[np.float64]:
    """

    Parameters
    ----------
    xy1 : np.ndarray
        Input coordinates.

    Returns
    -------
    np.ndarray
        Transformed input.
    """
    xy = np.ones_like(xy1, dtype=float)
    xy[:, 0] = 0.5 + (xy1[:, 0] - 0.5) * (1 - xy1[:, 1])
    xy[:, 1] = xy1[:, 1] * np.sqrt(3) / 2

    return xy


def make_mesh(ax: Axes) -> tuple[list[Any], list[Any]]:
    """Make coordinate mesh."""
    # Make coordinate mesh
    i = np.linspace(0, 1, 11).reshape(11, 1)
    text_handles: list[Text] = []
    line_handles: list[list[Line2D]] = []
    # 1
    xy1 = ternary_coord_trans(i, np.zeros(len(i)), 1 - i)
    xy2 = ternary_coord_trans(i, 1 - i, np.zeros(len(i)))
    for j in range(len(i)):
        line_handles.append(
            ax.plot(
                [
                    xy1[j, 0],
                    xy2[j, 0],
                ],
                [
                    xy1[j, 1],
                    xy2[j, 1],
                ],
                ":k",
                linewidth=0.25,
            )
        )
    # 2
    xy1 = ternary_coord_trans(np.zeros(len(i)), i, 1 - i)
    xy2 = ternary_coord_trans(1 - i, i, np.zeros(len(i)))
    for j in range(len(i)):
        line_handles.append(
            ax.plot(
                [
                    xy1[j, 0],
                    xy2[j, 0],
                ],
                [
                    xy1[j, 1],
                    xy2[j, 1],
                ],
                ":k",
                linewidth=0.25,
            )
        )
    # 3
    xy1 = ternary_coord_trans(1 - i, np.zeros(len(i)), i)
    xy2 = ternary_coord_trans(np.zeros(len(i)), 1 - i, i)
    for j in range(len(i)):
        line_handles.append(
            ax.plot(
                [
                    xy1[j, 0],
                    xy2[j, 0],
                ],
                [
                    xy1[j, 1],
                    xy2[j, 1],
                ],
                ":k",
                linewidth=0.25,
            )
        )
    # Tick mark text
    xy1 = ternary_coord_trans(1 - i, i, np.zeros(len(i)))
    for j in range(len(i)):
        text_handles.append(ax.text(xy1[j, 0], xy1[j, 1] - 0.025, "%.1f" % (i[j, 0])))
    xy1 = ternary_coord_trans(i, np.zeros(len(i)), 1 - i)
    for j in range(len(i)):
        text_handles.append(
            ax.text(xy1[j, 0] - 0.05, xy1[j, 1] + 0.025, "%.1f" % (i[j, 0]))
        )
    xy1 = ternary_coord_trans(np.zeros(len(i)), 1 - i, i)
    for j in range(len(i)):
        text_handles.append(ax.text(xy1[j, 0], xy1[j, 1] + 0.025, "%.1f" % (i[j, 0])))
    plt.setp(text_handles, fontsize=10)

    return text_handles, line_handles



from .carbonate_pressure_substitution import carbonate_pressure_model
from .parse_t_matrix_inputs import parse_t_matrix_inputs
from .run_t_matrix import run_t_matrix
from .t_matrix_C import t_matrix_porosity_c_alpha_v
from .t_matrix_opt_fluid_sub_exp import run_t_matrix_with_opt_params_exp
from .t_matrix_opt_fluid_sub_petec import run_t_matrix_with_opt_params_petec
from .t_matrix_opt_forward_model_exp import (
    run_t_matrix_forward_model_with_opt_params_exp,
)
from .t_matrix_opt_forward_model_min import (
    run_t_matrix_forward_model_with_opt_params_petec,
)
from .t_matrix_parameter_optimisation_exp import t_matrix_optimisation_exp
from .t_matrix_parameter_optimisation_min import t_matrix_optimisation_petec
from .t_matrix_vector import (
    array_inverse,
    array_matrix_mult,
    t_matrix_porosity_vectorised,
)

__all__ = [
    "carbonate_pressure_model",
    "parse_t_matrix_inputs",
    "run_t_matrix",
    "t_matrix_porosity_c_alpha_v",
    "run_t_matrix_with_opt_params_exp",
    "run_t_matrix_with_opt_params_petec",
    "run_t_matrix_forward_model_with_opt_params_exp",
    "run_t_matrix_forward_model_with_opt_params_petec",
    "t_matrix_optimisation_exp",
    "t_matrix_optimisation_petec",
    "array_inverse",
    "array_matrix_mult",
    "t_matrix_porosity_vectorised",
]



from typing import Any, cast

import numpy as np

from rock_physics_open.equinor_utilities import gen_utilities
from rock_physics_open.equinor_utilities.optimisation_utilities import (
    gen_opt_routine,
    opt_param_info,
    save_opt_params,
)
from rock_physics_open.equinor_utilities.optimisation_utilities.opt_subst_utilities import (
    opt_param_to_ascii,
)
from rock_physics_open.equinor_utilities.various_utilities.types import Array1D

from .curvefit_t_matrix_exp import curvefit_t_matrix_exp
from .t_matrix_parameter_optimisation_min import DEF_VP_VS_RATIO


def t_matrix_optimisation_exp(
    k_fl: Array1D[np.float64],
    rho_fl: Array1D[np.float64],
    por: Array1D[np.float64],
    vsh: Array1D[np.float64],
    vp: Array1D[np.float64],
    vs: Array1D[np.float64],
    rhob: Array1D[np.float64],
    angle: float = 0.0,
    k_r: float = 50.0,
    eta_f: float = 1.0,
    tau: float = 1.0e-7,
    freq: float = 1.0e3,
    file_out_str: str = "opt_params_exp.pkl",
    display_results: bool = False,
    well_name: str = "Unknown well",
    **opt_kwargs: Any,
) -> tuple[
    Array1D[np.float64],
    Array1D[np.float64],
    Array1D[np.float64],
    Array1D[np.float64],
    Array1D[np.float64],
    Array1D[np.float64],
    Array1D[np.float64],
    Array1D[np.float64],
]:
    """T-Matrix optimisation adapted to an exploration setting, where detailed well information is generally not known.
    Inclusion parameters are optimised for the whole well.

    Parameters
    ----------
    por : np.ndarray
        Inclusion porosity [ratio].
    vsh : np.ndarray
        Shale volume [ratio].
    k_fl : np.ndarray
        Fluid bulk modulus [Pa].
    rho_fl : np.ndarray
        Fluid density [kg/m^3].
    vp : np.ndarray
        Compressional velocity log [m/s].
    vs : np.ndarray
        Shear velocity log [m/s].
    rhob : np.ndarray
        Bulk density log [kg/m^3].
    angle : float
        Angle of symmetry plane [degrees]
    k_r : float
        Permeability [mD].
    eta_f : float
        Fluid viscosity [cP].
    tau : float
        Relaxation time constant [s].
    freq : float
        Signal frequency [Hz].
    file_out_str : str
        Output file name (string) to store optimal parameters (pickle format).
    display_results : bool
        Display optimal parameters in a window after run.
    well_name : str
        Name of well to be displayed in info box title.
    opt_kwargs : Any
        Additional keywords to be passed to optimisation function

    Returns
    -------
    tuple
        vp_mod, vs_mod - modelled logs, vp_res, vs_res - residual logs.
    """
    # 1. Preparation that is independent of search for minimum possible aspect ratio for second inclusion set

    # Optimisation function for selected parameters
    opt_fun = curvefit_t_matrix_exp
    # rho_min = (rhob - por * rho_fl) / (1 - por)
    # EXP adapted inputs: include fluid data and other params in x_data
    # expand single value parameters to match logs length
    por, angle_, k_r_, eta_f_, tau_, freq_, def_vpvs = cast(
        list[Array1D[np.float64]],
        gen_utilities.dim_check_vector(
            (por, angle, k_r, eta_f, tau, freq, DEF_VP_VS_RATIO)
        ),
    )
    x_data = np.stack(
        (por, vsh, k_fl, rho_fl, angle_, k_r_, eta_f_, tau_, freq_, def_vpvs), axis=1
    )
    # Set weight to vs to give vp and vs similar influence on optimisation
    y_data = np.stack([vp, vs * DEF_VP_VS_RATIO], axis=1)

    # 2. Search for minimum aspect ratio of inclusion set no. 2, given that inclusion set no. 1 will at least represent
    # 50% of inclusions. Minimum aspect ratio is linked to the porosity, and the most conservative estimate is to use
    # the maximum value. This is likely to deteriorate the optimisation results, so a search for a more appropriate
    # value that still produces valid results is made
    percentiles = [50, 75, 80, 85, 90, 95, 99, 100]
    valid_result = False
    vel_mod = None
    vel_res = None
    opt_params = None
    _, scale_val, _ = opt_param_info()
    while not valid_result and percentiles:
        try:
            # Make sure that parameters are not in conflict with T Matrix assumptions
            min_v1 = 0.5
            max_por = np.percentile(por, percentiles[0])
            min_a2 = (1.0 - min_v1) * max_por
            # Test with all parameters in the range 0.0 - 1.0
            # Params:      f_ani f_con a1 a2 v1 k_carb mu_carb rho_carb k_sh mu_sh rho_sh
            lower_bound = np.array(
                [
                    0.0,
                    0.0,
                    0.5,
                    min_a2,
                    min_v1,  # f_ani f_con a1 a2 v1
                    35.0 / scale_val["k_carb"],
                    30.0 / scale_val["mu_carb"],
                    2650.0 / scale_val["rho_carb"],  # k_carb, mu_carb, rho_carb
                    15.0 / scale_val["k_sh"],
                    7.0 / scale_val["mu_sh"],
                    2500.0 / scale_val["rho_sh"],
                ],
                dtype=float,
            )  # k_sh, mu_sh, rho_sh
            upper_bound = np.array(
                [
                    1.0,
                    1.0,
                    1.0,
                    0.30,
                    1.0,  # f_ani f_con a1 a2 v1
                    1.0,
                    1.0,
                    1.0,  # k_carb, mu_carb, rho_carb
                    1.0,
                    1.0,
                    1.0,
                ],
                dtype=float,
            )  # k_sh, mu_sh, rho_sh
            x0 = (upper_bound + lower_bound) / 2.0
            # Optimisation step without fluid substitution
            vel_mod, vel_res, opt_params = gen_opt_routine(
                opt_function=opt_fun,
                x_data_orig=x_data,
                y_data=y_data,
                x_init=x0,
                low_bound=lower_bound,
                high_bound=upper_bound,
                **opt_kwargs,
            )
            valid_result = True
        except ValueError:
            _ = percentiles.pop(0)
            valid_result = False

    if not valid_result:
        raise ValueError(
            f"{__file__}: unable to find stable value for T Matrix optimisation, second inclusion"
        )
    if vel_mod is None or vel_res is None or opt_params is None:
        raise ValueError(
            f"{__file__}: expected optimisation routine to return values, but got None"
        )

    # Reshape outputs and remove weight from vs
    vp_mod, vs_mod = [arr.flatten() for arr in np.split(vel_mod, 2, axis=1)]
    vp_res, vs_res = [arr.flatten() for arr in np.split(vel_res, 2, axis=1)]
    vs_mod = vs_mod / DEF_VP_VS_RATIO
    vs_res = vs_res / DEF_VP_VS_RATIO
    vpvs_mod = vp_mod / vs_mod
    # Calculate the modelled density
    rhob_mod = (
        (1.0 - vsh) * opt_params[7] * scale_val["rho_carb"]
        + vsh * opt_params[10] * scale_val["rho_sh"]
    ) * (1.0 - por) + por * rho_fl
    ai_mod = vp_mod * rhob_mod
    rhob_res = rhob_mod - rhob
    # Save the optimal parameters
    save_opt_params(
        opt_type="exp",
        opt_params=opt_params,
        file_name=file_out_str,
        well_name=well_name,
    )
    if display_results:
        opt_param_to_ascii(
            in_file=file_out_str,
            well_name=well_name,
        )

    return vp_mod, vs_mod, rhob_mod, ai_mod, vpvs_mod, vp_res, vs_res, rhob_res



from typing import Literal
from warnings import warn

import numpy as np
import numpy.typing as npt

from rock_physics_open.equinor_utilities import std_functions

from .constant_cement_models import constant_cement_model_dry
from .friable_models import CoordinateNumberFunction, friable_model_dry
from .patchy_cement_model import constant_cement_model_pcm

_BehaviorOptions = Literal["snap", "disregard"]


def patchy_cement_pressure_fluid_substitution(
    k_min: npt.NDArray[np.float64],
    mu_min: npt.NDArray[np.float64],
    rho_min: npt.NDArray[np.float64],
    k_cem: npt.NDArray[np.float64],
    mu_cem: npt.NDArray[np.float64],
    rho_cem: npt.NDArray[np.float64],
    k_fl_old: npt.NDArray[np.float64],
    rho_fl_old: npt.NDArray[np.float64],
    k_fl_new: npt.NDArray[np.float64],
    rho_fl_new: npt.NDArray[np.float64],
    phi: npt.NDArray[np.float64],
    p_eff_old: npt.NDArray[np.float64],
    p_eff_new: npt.NDArray[np.float64],
    vp_old: npt.NDArray[np.float64],
    vs_old: npt.NDArray[np.float64],
    rho_b_old: npt.NDArray[np.float64],
    p_eff_low: npt.NDArray[np.float64],
    frac_cem_up: npt.NDArray[np.float64] | float,
    frac_cem: npt.NDArray[np.float64] | float,
    shear_red: float,
    phi_c: float,
    coord_num_func: CoordinateNumberFunction,
    n: float,
    model_type: Literal["weight", "cement_fraction"] = "weight",
    phi_below_zero: _BehaviorOptions = "disregard",
    phi_above_phi_c: _BehaviorOptions = "snap",
    k_sat_above_k_min: _BehaviorOptions = "disregard",
    above_upper_bound: _BehaviorOptions = "snap",
    below_lower_bound: _BehaviorOptions = "disregard",
) -> tuple[
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
]:
    """
    Patchy cement model, developed by Per Avseth and Norunn Skjei. A sandstone model in which part of the
    matrix is pressure sensitive, and part is cemented.

    Control parameters with default value:
    model_type = 'weight': A sample by sample weight between cemented (upper) and friable (lower) bound are calculated.
        The alternative is 'cement_fraction', in which the weights are calculated from one constant cement model

    phi_below_zero='disregard': Values with negative porosity will either have their value snapped to zero ('snap')
    or the input value will be returned ('disregard'). These option should give the same
    result.

    phi_above_phi_c = 'disregard':	Values with porosity above critical porosity will either have their values snapped
    to critical porosity ('snap') or the input value will be returned ('disregard').

    k_sat_above_k_min='disregard': Values with saturated bulk modulus above the mineral bulk modulus will have their
    value snapped to the mineral modulus ('snap') or the input value will be returned ('disregard').

    above_upper_bound='snap': 	Values with moduli above upper bound will either be snapped to upper bound ('snap')
    or the input value will be returned ('disregard').

    below_lower_bound='disregard': Values with moduli below lower bound will either be snapped to lower bound ('snap')
    or the input value returned ('disregard').

    Comment
    -------
    Based on program by Norunn Skjei.
    Translated to Python by Harald Flesche.

    Parameters
    ----------
    k_min : np.ndarray
        Mineral bulk modulus [Pa].
    mu_min : np.ndarray
        Mineral shear modulus [Pa].
    rho_min : np.ndarray
        Mineral density [kg/m^3]
    k_cem : np.ndarray
        Sandstone cement bulk modulus [Pa].
    mu_cem : np.ndarray
        Sandstone cement shear modulus [Pa].
    rho_cem : np.ndarray
        Cement density [kg/m^3]
    k_fl_old : np.ndarray
        Initial fluid bulk modulus for sandstone fluid [Pa].
    rho_fl_old : np.ndarray
        Initial fluid bulk density for sandstone fluid [kg/m^3].
    k_fl_new : np.ndarray
        Substituted fluid bulk modulus for sandstone fluid [Pa].
    rho_fl_new : np.ndarray
        Substituted fluid bulk density for sandstone fluid [kg/m^3].
    phi : np.ndarray
        Total porosity [fraction].
    p_eff_old : np.ndarray
        Initial effective pressure [Pa].
    p_eff_new : np.ndarray
        Substituted effective pressure [Pa].
    p_eff_low : float
        Lower bound effective pressure [Pa].
    frac_cem_up : float
        Upper bound cement volume fraction [fraction].
    frac_cem : cement fraction of constant cement model - representative for observed data. Must be lower than
        frac_cem_up
    shear_red : float
        Shear reduction factor for sandstone [fraction].
    phi_c : float
        Critical porosity [fraction].
    n : float
        Coordination number [unitless].
    coord_num_func : str
        Indication if coordination number should be calculated from porosity or kept constant.
    vp_old : np.ndarray
        Initial p-wave velocity [m/s].
    vs_old : np.ndarray
        Initial s-wave velocity [m/s].
    rho_b_old : np.ndarray
        Initial bulk density [kg/m^3].
    model_type : str
        Version of model, either 'weight' or 'cement_fraction'
    phi_below_zero : str
        Control for handling of negative porosity samples.
    phi_above_phi_c : str
        Control for handling of porosity samples above critical porosity.
    k_sat_above_k_min : str
        Control for handling of bulk modulus samples above mineral bulk modulus.
    above_upper_bound : str
        Control for handling of samples above the upper bound.
    below_lower_bound : str
        Control for handling of samples below the lower bound.

    Returns
    -------
    tuple
        vp_new, vs_new, rho_b_new, ai_new, vpvs_new, k_sat_new, mu_new, wk, wmu, idx_valid :
        (np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray).
        vp_new :Saturated P-velocity [m/s] after fluid and pressure substitution,
        vs_new : Saturated S-velocity [m/s] after fluid and pressure substitution,
        rho_b_new : Saturated density [kg/m3] after fluid and pressure substitution,
        ai_new : Saturated acoustic impedance [kg/m3 x m/s] after fluid and pressure substitution,
        vpvs_new : Saturated Vp/Vs ratio [unitless] after fluid and pressure substitution,
        k_sat_new : New saturated rock bulk modulus [Pa],
        mu_new : New shear modulus [Pa],
        wk, wmu : weights for k and mu between upper and lower bound,
        idx_valid : samples fluid substitution is performed
    """

    # Original saturated bulk and shear modulus
    k_sat_old, mu_old = std_functions.moduli(vp=vp_old, vs=vs_old, rhob=rho_b_old)

    # Handling of samples that violate the assumptions of the Patchy Cement
    # substitution:
    #   1: phi <= 0
    #   2: phi > phi_c
    #   3: k_sat > k_min
    idx1 = _handle_exceptions_part_1(
        phi_vec=phi,
        phi_c_const=phi_c,
        k_sat=k_sat_old,
        mu=mu_old,
        k_min=k_min,
        mu_min=mu_min,
        phi_below_zero=phi_below_zero,
        phi_above_phi_c=phi_above_phi_c,
        k_sat_above_k_min=k_sat_above_k_min,
    )

    # k_dry for original pressure from Gassmann for valid samples
    k_dry_old = np.ones_like(phi) * np.nan
    k_dry_old[~idx1] = std_functions.gassmann_dry(
        k_sat=k_sat_old[~idx1],
        por=phi[~idx1],
        k_fl=k_fl_old[~idx1],
        k_min=k_min[~idx1],
    )

    # For common handling of zero-porosity point: calculate effective
    # properties using Hashin-Shtrikman
    k_zero, mu_zero = std_functions.hashin_shtrikman_walpole(
        k1=k_cem,
        mu1=mu_cem,
        k2=k_min,
        mu2=mu_min,
        f1=frac_cem_up,
        bound="lower",
    )

    # Lower bound for estimation of weight W (input moduli in Pa, pressure in Pa)
    k_low, mu_low = friable_model_dry(
        k_min=k_zero,
        mu_min=mu_zero,
        phi=phi,
        p_eff=p_eff_low * np.ones_like(phi),
        phi_c=phi_c,
        coord_num_func=coord_num_func,
        n=n,
        shear_red=shear_red,
    )

    # Upper bounds for estimation of weight W
    k_up, mu_up = constant_cement_model_dry(
        k_min=k_min,
        mu_min=mu_min,
        k_cem=k_cem,
        mu_cem=mu_cem,
        phi=phi,
        frac_cem=frac_cem_up,
        phi_c=phi_c,
        n=n,
        shear_red=shear_red,
        extrapolate_to_max_phi=False,
        return_k_zero=False,
    )

    # Handling of samples that violate the assumptions of the Patchy Cement
    # substitution:
    #   4: Model violation in Gassmann (k_dry < 0.0), friable model or constant cement
    #      model
    #   5: k_dry, mu below lower bound
    #   6: k_dry, mu above upper bound

    idx2 = _handle_exceptions_part_2(
        k_dry=k_dry_old,
        mu=mu_old,
        k_up=k_up,
        mu_up=mu_up,
        k_low=k_low,
        mu_low=mu_low,
        idx=idx1,
        above_upper_bound=above_upper_bound,
        below_lower_bound=below_lower_bound,
    )
    idx_valid = np.logical_not(idx2)

    # Pressure sensitive model for initial pressure
    k_dry_p_init, mu_p_init = friable_model_dry(
        k_min=k_zero,
        mu_min=mu_zero,
        phi=phi,
        p_eff=p_eff_old,
        phi_c=phi_c,
        coord_num_func=coord_num_func,
        n=n,
        shear_red=shear_red,
    )

    # Pressure sensitive model for final pressure
    k_dry_p_final, mu_p_final = friable_model_dry(
        k_min=k_zero,
        mu_min=mu_zero,
        phi=phi,
        p_eff=p_eff_new,
        phi_c=phi_c,
        coord_num_func=coord_num_func,
        n=n,
        shear_red=shear_red,
    )

    # Weights - either on a sample basis or by a representative constant cement model
    if model_type == "weight":
        w_k = _calculate_weight(
            dry_modulus=k_dry_old,
            low_modulus=k_low,
            high_modulus=k_up,
        )
        w_mu = _calculate_weight(
            dry_modulus=mu_old,
            low_modulus=mu_low,
            high_modulus=mu_up,
        )
    else:
        k_cc, mu_cc = constant_cement_model_pcm(
            kmin=k_min,
            mymin=mu_min,
            kcem=k_cem,
            mycem=mu_cem,
            kzero=k_zero,
            myzero=mu_zero,
            phi=phi,
            cem_frac=frac_cem,
            phic=phi_c,
            n=n,
            red_shear=shear_red,
        )
        w_k = _calculate_weight(
            dry_modulus=k_cc,
            low_modulus=k_low,
            high_modulus=k_up,
        )
        w_mu = _calculate_weight(
            dry_modulus=mu_cc,
            low_modulus=mu_low,
            high_modulus=mu_up,
        )

    w_cem = 0.5 * (w_mu + w_k)
    k_mod_dry = np.nan * np.ones_like(phi)
    k_mod = np.nan * np.ones_like(phi)
    mu_mod = np.nan * np.ones_like(phi)
    rho_mod = np.nan * np.ones_like(phi)

    # Calculate modelled values - first for the valid samples
    k_mod_dry[idx_valid] = k_dry_p_init[idx_valid] + w_k[idx_valid] * (
        k_up[idx_valid] - k_low[idx_valid]
    )
    k_mod[idx_valid] = std_functions.gassmann(
        k_dry=k_mod_dry[idx_valid],
        por=phi[idx_valid],
        k_fl=k_fl_old[idx_valid],
        k_min=k_min[idx_valid],
    )
    mu_mod[idx_valid] = mu_p_init[idx_valid] + w_mu[idx_valid] * (
        mu_up[idx_valid] - mu_low[idx_valid]
    )
    rho_mod[idx_valid] = (
        phi[idx_valid] * rho_fl_old[idx_valid]
        + (1.0 - phi[idx_valid] - w_cem[idx_valid] * frac_cem_up) * rho_min[idx_valid]
        + w_cem[idx_valid] * frac_cem_up * rho_cem[idx_valid]
    )

    # Fill in input values for invalid samples
    k_mod[idx2] = k_sat_old[idx2]
    mu_mod[idx2] = mu_old[idx2]
    rho_mod[idx2] = rho_b_old[idx2]
    vp_mod, vs_mod, _, _ = std_functions.velocity(
        k=k_mod,
        mu=mu_mod,
        rhob=rho_mod,
    )

    # Calculate residuals. For invalid samples, the residuals will be set to zero
    vp_res = vp_old - vp_mod
    vs_res = vs_old - vs_mod
    rho_res = rho_b_old - rho_mod

    # Estimate dry moduli values according to initial dry values, changed
    # pressure and estimated pressure sensitivity
    k_dry_new = (
        k_dry_old
        * (k_dry_p_final + w_k * (k_up - k_dry_p_final))
        / (k_dry_p_init + w_k * (k_up - k_dry_p_init))
    )
    mu_new = (
        mu_old
        * (mu_p_final + w_mu * (mu_up - mu_p_final))
        / (mu_p_init + w_mu * (mu_up - mu_p_init))
    )

    # New k_sat for new pressure from Gassmann
    k_sat_new = np.ones_like(phi) * np.nan
    if np.any(idx_valid):
        k_sat_new[idx_valid] = std_functions.gassmann(
            k_dry=k_dry_new[idx_valid],
            por=phi[idx_valid],
            k_fl=k_fl_new[idx_valid],
            k_min=k_min[idx_valid],
        )
    k_sat_new[idx2] = k_sat_old[idx2]
    mu_new[idx2] = mu_old[idx2]

    # New saturated density
    rho_b_new = np.ones_like(phi) * np.nan
    rho_b_new[idx_valid] = rho_b_old[idx_valid] + phi[idx_valid] * (
        rho_fl_new[idx_valid] - rho_fl_old[idx_valid]
    )
    rho_b_new[idx2] = rho_b_old[idx2]

    # New saturated velocities and derived values
    vp_new, vs_new, _, _ = std_functions.velocity(
        k=k_sat_new,
        mu=mu_new,
        rhob=rho_b_new,
    )
    ai_new = vp_new * rho_b_new
    vpvs_new = vp_new / vs_new

    return (
        vp_new,
        vs_new,
        rho_b_new,
        ai_new,
        vpvs_new,
        w_k,
        w_mu,
        idx_valid,
        vp_res,
        vs_res,
        rho_res,
    )


def _handle_exceptions_part_1(
    phi_vec: npt.NDArray[np.float64],
    phi_c_const: float,
    k_sat: npt.NDArray[np.float64],
    mu: npt.NDArray[np.float64],
    k_min: npt.NDArray[np.float64],
    mu_min: npt.NDArray[np.float64],
    phi_below_zero: _BehaviorOptions = "disregard",
    phi_above_phi_c: _BehaviorOptions = "snap",
    k_sat_above_k_min: _BehaviorOptions = "disregard",
) -> np.ndarray:
    # Handling of samples that violate assumptions of the Patchy Cement substitution:
    #   1: phi < 0
    #   2: phi > phi_c
    #   3: k_sat > k_min, mu > mu_min

    # Handling of case 1:
    idx1 = phi_vec < 0.0
    if phi_below_zero == "snap":
        phi_vec[idx1] = 0.0
        idx1 = np.zeros(phi_vec.shape).astype(bool)
    elif phi_below_zero == "disregard":
        pass
    else:
        raise ValueError('unknown argument for parameter "phi_below_zero"')  # pyright: ignore[reportUnreachable] | Kept for backward compatibility

    #  Handling of case 2:
    idx2 = phi_vec > phi_c_const
    if phi_above_phi_c == "snap":
        phi_vec[idx2] = phi_c_const
        idx2 = np.zeros(phi_vec.shape).astype(bool)
    elif phi_above_phi_c == "disregard":
        pass
    else:
        raise ValueError('unknown argument for parameter "phi_above_phi_c"')  # pyright: ignore[reportUnreachable] | Kept for backward compatibility

    # Handling of case 3:
    idx3 = np.logical_or(k_sat > k_min, mu > mu_min)
    if k_sat_above_k_min == "snap":
        k_min[idx3] = k_sat[idx3]
        mu_min[idx3] = mu[idx3]
        idx3 = np.zeros(k_sat.shape).astype(bool)
    elif k_sat_above_k_min == "disregard":
        pass
    else:
        raise ValueError('unknown argument for parameter "k_sat_above_k_min"')  # pyright: ignore[reportUnreachable] | Kept for backward compatibility
    return np.any(np.vstack((idx1, idx2, idx3)), axis=0)


def _handle_exceptions_part_2(
    k_dry: np.ndarray,
    mu: np.ndarray,
    k_up: np.ndarray,
    mu_up: np.ndarray,
    k_low: np.ndarray,
    mu_low: np.ndarray,
    idx: np.ndarray,
    above_upper_bound: _BehaviorOptions = "snap",
    below_lower_bound: _BehaviorOptions = "disregard",
) -> np.ndarray:
    # Handling of samples that violate the assumptions of the Patchy Cement
    # substitution:
    #   4: Model violation in Gassmann (k_dry < 0.0), friable model or constant cement
    #       model
    #   5: k_dry, mu below lower bound
    #   6: k_dry, mu above upper bound

    # Identification of case 4 and calculation of dry rock properties:
    # gassmann_dry returns dry property < 0 as NaN
    idx_nan = np.isnan(k_dry)

    # Can be non-physical cases in friable model
    idx_nan = np.logical_or(idx_nan, np.isnan(k_low))

    # Can be non-physical cases in constant cement model
    idx_nan = np.logical_or(idx_nan, np.isnan(k_up))

    # Handling of case 5:
    idx5k = np.zeros_like(k_dry).astype(bool)
    idx5mu = np.zeros_like(k_dry).astype(bool)
    _ = np.less(k_dry, k_low, out=idx5k, where=~idx_nan)
    _ = np.less(mu, mu_low, out=idx5mu, where=~idx_nan)
    if below_lower_bound == "snap":
        k_dry[idx5k] = k_low[idx5k]
        mu[idx5mu] = mu_low[idx5mu]
        idx5 = np.zeros(k_dry.shape).astype(bool)
    elif below_lower_bound == "disregard":
        idx5 = np.logical_or(idx5k, idx5mu)
    else:
        raise ValueError('unknown argument for parameter "below_lower_bound"')  # pyright: ignore[reportUnreachable] | Kept for backward compatibility

    # Handling of case 6:
    if above_upper_bound == "snap":
        idx6k = np.zeros_like(k_dry).astype(bool)
        idx6mu = np.zeros_like(k_dry).astype(bool)
        _ = np.greater(k_dry, k_up, out=idx6k, where=~idx_nan)
        _ = np.greater(mu, mu_up, out=idx6mu, where=~idx_nan)
        k_dry[idx6k] = k_up[idx6k]
        mu[idx6mu] = mu_up[idx6mu]
        idx6 = np.zeros(k_dry.shape).astype(bool)
    elif above_upper_bound == "disregard":
        idx6 = np.logical_or(
            np.greater(k_dry, k_up, where=~idx_nan),
            np.greater(mu, mu_up, where=~idx_nan),
        )
    else:
        raise ValueError('unknown argument for parameter "above_upper_bound"')  # pyright: ignore[reportUnreachable] | Kept for backward compatibility

    # Exception for all cases 1-6:
    return np.any(np.vstack((idx, idx5, idx6, idx_nan)), axis=0)


def _calculate_weight(
    dry_modulus: npt.NDArray[np.float64],
    low_modulus: npt.NDArray[np.float64],
    high_modulus: npt.NDArray[np.float64],
    force_full_range: bool = False,
) -> npt.NDArray[np.float64]:
    """
    Calculates a weight for a value between a lower and upper bound. Used for moduli


    Parameters
    ----------
    dry_modulus: value to be evaluated
    low_modulus: lower bound
    high_modulus: upper bound
    force_full_range: do not clip weight to a range of [0.0, 1.0]

    Returns
    -------
    weight: value between [0.0, 1.0]
    """
    idx = np.abs(high_modulus - low_modulus) < 2.0 * np.finfo(float).eps
    if np.any(idx):
        warn(
            f"weight estimation: high and low bound is identical for {np.sum(high_modulus == low_modulus)} samples"
        )
    weight = np.ones_like(dry_modulus)
    weight[~idx] = (dry_modulus[~idx] - low_modulus[~idx]) / (
        high_modulus[~idx] - low_modulus[~idx]
    )
    # Catch cases outside the allowed range
    if not force_full_range:
        weight = np.clip(weight, 0.0, 1.0)
    return weight



from typing import cast

import numpy as np
import numpy.typing as npt

from rock_physics_open.equinor_utilities.gen_utilities import dim_check_vector

from .pq import p_q_fcn


def multi_sca(
    *args: npt.NDArray[np.float64],
    tol: float,
) -> tuple[
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
    npt.NDArray[np.float64],
]:
    """SCA - Effective elastic moduli using Berryman's Self-Consistent
    (Coherent Potential Approximation) method.

    -----------------------------------------------

    Inputs for each phase:

    k:      mineral bulk modulus [Pa]

    mu:     mineral shear modulus [Pa]

    rho:    mineral density [kg/m^3]

    asp:    aspect ratio for inclusions [ratio]

    frac:   fraction of matrix that is made up of this mineral [fraction]

    Additional inputs:

    tol:    tolerance parameter (scalar) [unitless]

    The sum of all fractions must add to one, and it will be normalised to one if it differs.

    =======================================================

    Based on function by T. Mukerji, SRB, Stanford University, 1994.

    Ported to Python by Harald Flesche, Equinor 2015.

    Parameters
    ----------
    args : list or tuple
        List or tuple containing multiples of elastic properties as explained above, all numpy arrays.
    tol : float
        Tolerance for the SCA iterations.

    Returns
    -------
    tuple
        k_sc, mu_sc, rhob : (np.ndarray, np.ndarray, np.ndarray).
        effective medium properties k_sc: bulk modulus [Pa], mu_sc: shear modulus [Pa], rhob: bulk density [kg/m^3].
    """
    if len(args) % 5 != 0:
        raise ValueError(
            "Call function with (k_sc,mu_sc,rhob) = multi_sca((k1,mu1,rho1,asp1,frac1, ...,k_n,mu_n,rho_n,asp_n,frac_n,tol)). Fractions must add up to 1.0"
        )

    args_ = cast(list[npt.NDArray[np.float64]], dim_check_vector(args))

    # Proceed without the parameter, all other inputs should be arrays

    n_mins = len(args_) // 5

    # Sort inputs
    k_min: list[npt.NDArray[np.float64]] = []
    mu_min: list[npt.NDArray[np.float64]] = []
    rho_min: list[npt.NDArray[np.float64]] = []
    frac: list[npt.NDArray[np.float64]] = []
    asp: list[npt.NDArray[np.float64]] = []

    for i in range(n_mins):
        k_min.append(args_[5 * i])
        mu_min.append(args_[5 * i + 1])
        rho_min.append(args_[5 * i + 2])
        asp.append(args_[5 * i + 3])
        frac.append(args_[5 * i + 4])

    # Check and normalise frac
    tot_frac = np.sum(np.vstack(frac[:]), 0)
    if np.any(tot_frac != 1.0):
        for i in range(n_mins):
            frac[i] = frac[i] / tot_frac

    # Initiate k_sc and mu_sc with weighted sum of phases
    k_sc = np.sum(np.vstack(k_min[:]) * np.vstack(frac[:]), 0)
    mu_sc = np.sum(np.vstack(mu_min[:]) * np.vstack(frac[:]), 0)
    rhob = np.sum(np.vstack(rho_min[:]) * np.vstack(frac[:]), 0)

    idx = np.any(np.vstack(frac[:]) == 1.0, 0)
    if np.any(idx):
        # If any of the samples contain 100% of one phase - substitute the result
        # with the mineral properties of that phase
        for i in range(n_mins):
            idx1 = frac[i] == 1.0
            if np.any(idx1):
                k_sc[idx1] = k_min[i][idx1]
                mu_sc[idx1] = mu_min[i][idx1]
            # Continue with the real mixtures
            k_min[i] = k_min[i][~idx]
            mu_min[i] = mu_min[i][~idx]
            asp[i] = asp[i][~idx]
            frac[i] = frac[i][~idx]
    if not np.all(idx):
        niter = 0
        # Initiate the delta with a value that makes the while-loop run
        delta = k_sc[~idx]
        # Express tolerance in terms of k0
        tol_ = tol * k_min[0]

        # Iterate until delta is less than the tolerance and the loop has run for
        # less than 3000 iterations (normally far less)
        while np.any(np.logical_and(np.greater(delta, tol_), np.less(niter, 3000))):
            p: list[npt.NDArray[np.float64]] = []
            q: list[npt.NDArray[np.float64]] = []
            for i in range(n_mins):
                p_tmp, q_tmp = p_q_fcn(
                    k_sc[~idx], mu_sc[~idx], k_min[i], mu_min[i], asp[i]
                )
                p.append(p_tmp)
                q.append(q_tmp)
            k_new = np.sum(
                np.vstack(frac[:]) * np.vstack(k_min[:]) * np.vstack(p[:]), 0
            ) / np.sum(np.vstack(frac[:]) * np.vstack(p[:]), 0)
            mu_new = np.sum(
                np.vstack(frac[:]) * np.vstack(mu_min[:]) * np.vstack(q[:]), 0
            ) / np.sum(np.vstack(frac[:]) * np.vstack(q[:]), 0)

            delta = np.absolute(k_sc[~idx] - k_new)
            k_sc[~idx] = k_new
            mu_sc[~idx] = mu_new

            niter = niter + 1

    return k_sc, mu_sc, rhob



from .dem import dem_model
from .dem_dual_por import dem_model_dual_por
from .kus_tok import kuster_toksoz_model
from .multi_sca import multi_sca
from .pq import p_q_fcn
from .sca import self_consistent_approximation_model
from .shale4_mineral import shale_model_4_mineral_dem
from .shale4_mineral_dem_overlay import shale_4_min_dem_overlay

__all__ = [
    "dem_model",
    "dem_model_dual_por",
    "kuster_toksoz_model",
    "multi_sca",
    "p_q_fcn",
    "self_consistent_approximation_model",
    "shale_model_4_mineral_dem",
    "shale_4_min_dem_overlay",
]
